
/* Smart UI v1.0.2 (2021-08-17) 
Copyright (c) 2011-2021 jQWidgets. 
License: https://htmlelements.com/license/ */ //

class DataAdapter {
    constructor(config) {
        if (!config) {
            config = {};
        }

        const that = Object.assign(this, config);

        const generateKey = function () {
            const S4 = function () {
                return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
            };
            return S4();
        };

        that.key = generateKey();

        that.$document = Smart.Utilities.Extend(document);
        that.boundSource = false === that.observable || that.virtualDataSource ? [] : new Smart.ObservableArray();
        that.dataItemById = [];

        if (that.allowAdd === undefined) {
            that.allowAdd = true;
        }

        if (that.allowRemove === undefined) {
            that.allowRemove = true;
        }

        if (that.allowUpdate === undefined) {
            that.allowUpdate = true;
        }

        if (that.mapChar === undefined) {
            that.mapChar = '.';
        }

        if (config.observable === undefined && !that.virtualDataSource) {
            that.observable = true;
        }
        else {
            that.observable = false;
        }

        if (that.sanitizeHTML === undefined) {
            that.sanitizeHTML = 'blackList';
        }

        that._expandedRowIds = [];

        if (!config.dataSource) {
            that.dataSource = [];
        }

        if (!config.dataFields) {
            that.dataFields = [];
        }

        if (!config.dataSourceType) {
            that.dataSourceType = 'array';
        }

        if (!config.id) {
            that.id = null;
        }

        if (!config.autoFetch) {
            that.autoFetch = true;
        }

        if (config.dataFields) {
            that.dataFields = config.dataFields;
        }

        Object.defineProperty(that, 'groupBy', {
            configurable: false,
            enumerable: true,
            get() {
                if (!that._groupBy) {
                    return [];
                }

                return that._groupBy;
            },
            set(value) {
                const updateGrouping = () => {
                    that.boundHierarchy = null;
                    that.refreshHierarchy();

                    if (that.onGroup) {
                        that.onGroup();
                    }
                }

                that._groupBy = new Smart.ObservableArray(value);
                that._groupBy.notify(function () {
                    updateGrouping();
                });

                if (that.isInitialized) {
                    updateGrouping();
                }
            }
        });

        if (!config.groupBy) {
            that.groupBy = [];
        }
        else {
            if (config.groupBy.toArray) {
                that.groupBy = config.groupBy.toArray();
            }
            else {
                that.groupBy = config.groupBy;
            }
        }

        if (config && config.autoBind !== false) {
            that.dataBind();
        }

        that.isInitialized = true;
    }

    beginUpdate() {
        const that = this;

        that.isUpdating = true;
    }

    endUpdate() {
        const that = this;

        that.isUpdating = false;
        that.refreshIndexes();
        that.refreshHierarchy();
    }

    get expandHierarchy() {
        if (this._expandHierarchy === undefined) {
            this._expandHierarchy = false;
        }

        return this._expandHierarchy;
    }

    set expandHierarchy(value) {
        const that = this;

        that._expandHierarchy = value;

        if (that.isInitialized) {
            that.refreshHierarchy();
        }
    }

    _getDataField(dataField) {
        switch (dataField.name) {
            case 'leaf':
            case 'parent':
            case 'expanded':
            case 'checked':
            case 'selected':
            case 'level':
            case 'icon':
            case 'data':
            case 'group':
            case 'children':
            case 'items': {
                if (dataField.map) {
                    dataField.map = dataField.map.replace(dataField.name, '$' + dataField.name);
                }
                else {
                    dataField.map = dataField.name;
                }
                dataField.name = '$' + dataField.name;
            }
        }
    }

    get dataFields() {
        const that = this;

        return that._dataFields;
    }

    set dataFields(value) {
        const that = this;

        that._dataFields = that._getDataFieldObjects(value);

        return that._dataFields;
    }

    _getDataFieldObjects(dataFields) {
        const that = this;

        let dataFieldObjects = [];

        if (typeof dataFields === 'number') {
            const charCode = 'A'.charCodeAt(0);
            let prefix = '';
            let index = 0;

            for (let i = 0; i < dataFields; i++) {
                const letter = String.fromCharCode(charCode + index);

                index++;

                const label = prefix + letter;

                dataFieldObjects.push({ name: label, dataType: 'string' })

                if (index >= 26) {
                    index = 0;
                    prefix += 'A';
                }
            }
        }
        else if (Array.isArray(dataFields) && dataFields.length > 0) {
            for (let i = 0; i < dataFields.length; i++) {
                const dataField = dataFields[i];

                if (typeof dataField === 'string') {
                    const dataFieldParts = dataField.split(':');
                    const name = dataFieldParts[0].trim();
                    const dataType = dataFieldParts.length > 1 ? dataFieldParts[1].trim() : 'string';

                    dataFieldObjects.push({ name: name, dataType: dataType });
                }
                else {
                    dataFieldObjects.push(dataField);
                }

                that._getDataField(dataFieldObjects[dataFieldObjects.length - 1]);
            }
        }

        return dataFieldObjects;
    }

    get dataSource() {
        const that = this;

        if (!that._dataSource) {
            that._dataSource = [];
        }

        return that._dataSource;
    }

    set dataSource(value) {
        const that = this;

        that._dataSource = value;

        if (that.isInitialized) {
            that.boundSource = false === that.observable || that.virtualDataSource ? [] : new Smart.ObservableArray();
            that.dataItemById = [];
            that.bindingCompleted = false;
            that.dataBind();
            that.refreshIndexes();
        }
    }

    get canNotify() {
        const that = this;

        if (that._canNotify === undefined) {
            that._canNotify = true;
        }

        return that._canNotify;
    }

    set canNotify(value) {
        const that = this;

        that._canNotify = value;
    }

    _notify(changeArgs) {
        const that = this;

        if (!that.canNotify) {
            return;
        }

        if (that.notifyFn) {
            that.notifyFn(changeArgs);
        }
    }

    notify(notifyFn) {
        const that = this;

        if (notifyFn) {
            that.notifyFn = notifyFn;
        }
    }

    toArray() {
        const that = this;

        if (false === that.observable) {
            return that.boundSource;
        }

        return that.boundSource.toArray();
    }

    dataBind() {
        const that = this;

        that.bindingCompleted = false;
        that.clear();

        const completed = () => {
            if (that.observable) {
                if (!that.boundSource.notify) {
                    that.boundSource = new Smart.ObservableArray();
                }

                that.boundSource.notify(function (event) {
                    if (event.action === 'update' && event.path && event.path.indexOf('.') >= 0
                        && event.path.indexOf('children') === -1
                        && event.path.indexOf('loaded') === -1
                        && event.path.indexOf('level') === -1
                        && event.path.indexOf('leaf') === -1
                        && event.path.indexOf('expanded') === -1) {

                        let canNotify = false;

                        for (let i = 0; i < that.dataFields.length; i++) {
                            const dataField = that.dataFields[i].name;

                            if (event.path.indexOf(dataField) >= 0) {
                                canNotify = true;
                            }
                        }

                        if (that._notify && canNotify) {
                            that._notify({ action: 'update', data: event.target, index: event.index });
                        }
                    }
                });
            }

            that._onBindingComplete();
        }

        if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('.json') >= 0 || that.dataSourceType === 'json')) {
            that.url = that.dataSource;
            that.dataSourceType = 'json';

            new Ajax(that, (data/*, status*/) => {
                that.dataSource = data;
            });
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('.xml') >= 0)) {
            that.url = that.dataSource;
            that.dataSourceType = 'xml';

            new Ajax(that, (data/*, status*/) => {
                that.dataSource = data;
            });
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('.xlsx') >= 0)) {
            that.url = that.dataSource;
            that.dataSourceType = 'xlsx';

            new Ajax(that, (data/*, status*/) => {
                if (!data[0]) {
                    data = [];
                    that._bindToArray();
                    completed();
                    return;
                }

                const keys = Object.keys(data[0]);
                const dataFieldMap = {};
                const dataRows = [];

                if (that.exportHeader !== false) {
                    let index = 0;

                    for (let key in keys) {
                        const name = keys[key];

                        dataFieldMap[name] = that.dataFields[index++].name;
                    }

                    for (let i = 1; i < data.length; i++) {
                        const row = data[i];
                        const dataRow = {};

                        for (let key in keys) {
                            const name = keys[key];

                            dataRow[dataFieldMap[name]] = row[name];
                        }

                        dataRows.push(dataRow);
                    }

                    that.dataSource = dataRows;
                }

                that._bindToArray();
                completed();
            });
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('.csv') >= 0)) {
            that.url = that.dataSource;
            that.dataSourceType = 'csv';

            new Ajax(that, (data) => {
                that.dataSource = data;
            });
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('.tsv') >= 0)) {
            that.url = that.dataSource;
            that.dataSourceType = 'tsv';

            new Ajax(that, (data) => {
                that.dataSource = data;
            });
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('.ics') >= 0 || (that.dataSource.indexOf('.php') >= 0 && that.dataSourceType === 'ics') || (that.dataSource.indexOf('.txt') >= 0 && that.dataSourceType === 'ics'))) {
            that.url = that.dataSource;
            that.dataSourceType = 'ics';

            new Ajax(that, (data) => {
                that.dataSource = data;
            });
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('BEGIN:VCALENDAR') >= 0 || that.dataSourceType === 'ics')) {
            that._bindToICS();
            completed();
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('<?xml') >= 0 || that.dataSourceType === 'xml')) {
            that._bindToXML();
            completed();
        }
        else if (typeof that.dataSource === 'string' && (that.dataSource.indexOf('  ') >= 0 || that.dataSource.indexOf(',') >= 0 || that.dataSourceType === 'tsv' || that.dataSourceType === 'csv' || that.dataSourceType === 'tab')) {
            that._bindToCSV();
            completed();
        }
        else if (that.dataSourceType === 'array') {
            that._bindToArray();
            completed();
        }
        else if (that.dataSourceType === 'json') {
            that._bindToJSON();
            completed();
        }
    }

    _bindToXML() {
        const that = this;

        let parser = null;
        let xmlDoc = null;

        if (window.DOMParser) {
            parser = new DOMParser();
            xmlDoc = parser.parseFromString(that.dataSource, 'text/xml');
        }

        if (that.root === undefined) that.root = '';
        if (that.record === undefined) that.record = '';

        const xmlRecords = xmlDoc.querySelectorAll(that.root + ' ' + that.record);
        const length = xmlRecords.length;
        const records = [];

        let dataFieldsLength = that.dataFields ? that.dataFields.length : 0;
        // auto-generate data columns
        if (dataFieldsLength === 0) {
            const firstRecord = xmlRecords[0];
            const dataFields = new Array();
            for (let obj in firstRecord) {
                const field = obj;
                dataFields[dataFields.length] = { name: field };
            }
            that.dataFields = dataFields;

            dataFieldsLength = dataFields.length;
        }

        for (let i = 0; i < length; i++) {
            const record = xmlRecords[i];
            const dataRow = {};

            if (record === undefined) {
                break;
            }

            if (that.id) {
                const id = record.querySelector(that.id);
                dataRow[that.id] = id.textContent;
            }

            for (let j = 0; j < dataFieldsLength; j++) {
                const dataField = that.dataFields[j];
                const value = record.querySelector(dataField.map || dataField.name);

                dataRow[dataField.name] = value.textContent;
            }

            records[i] = dataRow;
        }

        that.canNotify = false;
        that.dataSource = records;
        that.canNotify = true;

        that._bindToArray();
    }

    _bindToCSV() {
        const that = this;

        const data = '' + that.dataSource;
        const rowDelimiter = '\n';
        const columnDelimiter = data.indexOf(',') === -1 ? '\t' : ',';
        const rowsWithNewRowDelimiter = data.split('\r');
        const records = new Array();

        let rows = data.split(rowDelimiter);
        let length = rows.length;

        if (length === 1 && rowsWithNewRowDelimiter.length > 1) {
            rows = rowsWithNewRowDelimiter;
            length = rows.length;
        }

        const hasColumns = that.dataFields.length > 0;

        for (let i = 0; i < length; i++) {
            const record = rows[i];
            const row = {};
            const columns = rows[i].split(columnDelimiter);

            for (let j = 0; j < columns.length; j++) {
                let dataField = that.dataFields[j];
                let value = columns[j];

                if (!dataField) {
                    that.dataFields.push({ name: value.trim() })
                    dataField = that.dataFields[that.dataFields.length - 1];
                }

                if (dataField.map) {
                    value = record[dataField.map];
                }

                value = value.trim();

                // skips columns row.
                if (i === 1 && !hasColumns) {
                    let dataType = 'string';

                    if (value === 'false' || value === 'true') {
                        dataType = 'boolean';
                        value = value === 'false' ? false : true;
                    }
                    else if (!isNaN(parseInt(value) && value.indexOf('/') === -1)) {
                        dataType = 'number';
                        value = parseFloat(value);
                    }
                    else if ((value.indexOf('-') >= 0 || value.indexOf('/') >= 0 || value.indexOf(':') >= 0) && new Date(value).toString() !== 'Invalid Date') {
                        dataType = 'date';
                        value = new Date(value);
                    }

                    if (!dataField.dataType) {
                        dataField.dataType = dataType;
                    }
                }
                else {
                    switch (dataField.dataType) {
                        case 'boolean':
                            value = value === 'false' ? false : true;
                            break;
                        case 'number':
                            value = parseFloat(value);
                            break;
                        case 'date':
                            value = new Date(value);
                            break;
                    }
                }

                const key = dataField.map || dataField.name || j.toString();
                row[key] = value;
            }

            if (hasColumns) {
                records.push(row);
            }
            else if (i > 0) {
                records.push(row);
            }
        }

        that.canNotify = false;
        that.dataSource = records;
        that.canNotify = true;

        that._bindToArray();
    }

    _onBindingComplete() {
        const that = this;

        that._buildHierarchy();

        if (that.onBindingComplete) {
            that.onBindingComplete({ data: that.boundSource });
        }

        if (that._notify) {
            that._notify({ action: 'bindingComplete', data: that.boundSource });
        }

        that.bindingCompleted = true;
    }

    refreshHierarchy() {
        const that = this;

        if (that.isUpdating) {
            return;
        }

        that._buildHierarchy();
    }

    find() {
        const that = this;

        return that.boundSource.find.apply(that.boundSource, arguments);
    }

    onVirtualDataSourceRequested(requestCallback, details) {
        const that = this;

        let first = details ? details.first : Infinity;
        let last = details ? details.last : Infinity;
        let row = details ? details.row : null;

        if (undefined === first) {
            first = Infinity;
        }

        if (undefined === last) {
            last = Infinity;
        }

        that.virtualFirstIndex = first;
        that.virtualLastIndex = last;

        if (that.virtualDataSource) {
            const getDataSource = function (dataAdapterSettings) {
                if (dataAdapterSettings.virtualDataSourceLength !== undefined) {
                    that.virtualDataSourceLength = dataAdapterSettings.virtualDataSourceLength;
                }

                new Smart.DataAdapter(
                    {
                        dataSource: dataAdapterSettings.dataSource,
                        dataFields: dataAdapterSettings.dataFields || that.dataFields,
                        data: details,
                        keyDataField: dataAdapterSettings.keyDataField,
                        parentDataField: dataAdapterSettings.parentDataField,
                        observable: false,
                        id: dataAdapterSettings.id || that.id,
                        onBindingComplete(event) {
                            if (details.result === undefined) {
                                details.result = true;
                            }

                            if (that.virtualDataSourceOnExpand && row) {
                                if (that.groupBy.length > 0) {
                                    if (event.data && event.data.length > 0) {
                                        that.add(event.data);
                                    }
                                }
                                else {
                                    if (event.data && event.data.length > 0) {
                                        that.add(event.data, row.$.id);
                                    }
                                    else {
                                        row.leaf = true;
                                    }
                                }

                                if (that.onFilter) {
                                    that.onFilter()
                                }

                                requestCallback(details);

                                return;
                            }

                            if (first === Infinity && details.action === 'scroll') {
                                if (event.data && event.data.length) {
                                    that.add(event.data);
                                }
                            }
                            else {
                                if (details.action === 'filter' || details.action === 'group') {
                                    that.canNotify = false;
                                    that.dataSource = event.data;
                                    that.canNotify = true;
                                }
                                else if (details.action === 'add' || details.action === 'update' || details.action === 'remove') {
                                    that.canNotify = false;
                                    that.dataSource = event.data;
                                    that.canNotify = true;
                                }
                                else {
                                    that.canNotify = false;
                                    that.dataSource = event.data;
                                    that.canNotify = true;
                                }
                            }


                            if (that.onFilter) {
                                that.onFilter()
                            }

                            requestCallback(details);
                        }
                    });
            }

            // const isEmpty = ( obj ) => Object.entries( obj ).length === 0 && ( obj.constructor === Object || obj.constructor === Array );

            let sqlQuery = {};

            if (details.sorting && details.sorting.length > 0) {
                let sorting = [];
                for (let dataField in details.sorting) {
                    sorting.push(dataField + ' ' + details.sorting[dataField].sortOrder.toUpperCase());
                }

                const orderByQuery = ' ORDER BY ' + sorting.join(', ');

                sqlQuery['orderBy'] = orderByQuery;
            }
            else {
                sqlQuery['orderBy'] = '';
            }

            if (first !== undefined && last !== undefined && !(first === 0 && last === 0)) {
                const blockSize = last - first;
                const limitQuery = ' LIMIT ' + blockSize + ' OFFSET ' + first;

                sqlQuery['limit'] = limitQuery;

                if (isNaN(blockSize)) {
                    sqlQuery['limit'] = '';
                }
            }
            else {
                sqlQuery['limit'] = '';
            }

            if (details.grouping && details.grouping.length > 0) {
                const groupByQuery = ' GROUP BY ' + details.grouping[0];

                sqlQuery['groupBy'] = groupByQuery;
            }
            else {
                sqlQuery['groupBy'] = '';
            }


            if (details.filtering && details.filtering.length > 0) {

                const textFilterMapper = (key, item) => {
                    switch (item.condition) {
                        case 'EMPTY':
                            return key + ' = ' + '\'\'';
                        case 'NOT_EMPTY':
                            return key + ' != ' + '\'\'';
                        case 'EQUAL':
                            return key + ' = \'' + item.value + '\'';
                        case 'NOT_EQUAL':
                            return key + ' != \'' + item.value + '\'';
                        case 'CONTAINS_CASE_SENSITIVE':
                        case 'CONTAINS':
                            return key + ' LIKE \'%' + item.value + '%\'';
                        case 'DOES_NOT_CONTAIN_CASE_SENSITIVE':
                        case 'DOES_NOT_CONTAIN':
                            return key + ' NOT LIKE \'%' + item.value + '%\'';
                        case 'STARTS_WITH_CASE_SENSITIVE':
                        case 'STARTS_WITH':
                            return key + ' LIKE \'' + item.value + '%\'';
                        case 'ENDS_WITH_CASE_SENSITIVE':
                        case 'ENDS_WITH':
                            return key + ' LIKE \'%' + item.value + '\'';
                        case 'NULL':
                            return key + ' IS NULL';
                        case 'NOT_NULL':
                            return key + ' IS NOT NULL';
                        default:
                            break
                    }
                }

                const numberFilterMapper = (key, item) => {
                    switch (item.condition) {
                        case 'EQUAL':
                            return key + ' = ' + item.value;
                        case 'NOT_EQUAL':
                            return key + ' != ' + item.value;
                        case 'GREATER_THAN':
                            return key + ' > ' + item.value;
                        case 'GREATER_THAN_OR_EQUAL':
                            return key + ' >= ' + item.value;
                        case 'LESS_THAN':
                            return key + ' < ' + item.value;
                        case 'LESS_THAN_OR_EQUAL':
                            return key + ' <= ' + item.value;
                        case 'NULL':
                            return key + ' IS NULL';
                        case 'NOT_NULL':
                            return key + ' IS NOT NULL';
                        case 'RANGE':
                            return (
                                '(' +
                                key +
                                ' >= ' +
                                item.value +
                                ' and ' +
                                key +
                                ' <= ' +
                                item.valueTo +
                                ')'
                            );
                        default:
                            break;
                    }
                }

                const createFilterSql = (mapper, key, item) => {
                    if (item.operator) {
                        var condition1 = mapper(key, item.condition1);
                        var condition2 = mapper(key, item.condition2);

                        return '(' + condition1 + ' ' + item.operator + ' ' + condition2 + ')';
                    }

                    return mapper(key, item);
                }

                let whereParts = [];

                for (let dataField in details.filtering) {
                    const filter = details.filtering[dataField];
                    let filters = filter.filters;
                    let columnWhereParts = [];
                    let columnOperator = ' AND ';

                    for (let i = 0; i < filters.length; i++) {
                        const item = filters[i];

                        switch (item.type) {
                            case 'stringFilter':
                                columnWhereParts.push(createFilterSql(textFilterMapper, dataField, item));
                                break;
                            case 'numericFilter':
                            case 'dateFilter':
                                columnWhereParts.push(createFilterSql(numberFilterMapper, dataField, item));
                                break;
                            case 'booleanFilter':
                                columnWhereParts.push(createFilterSql(numberFilterMapper, dataField, item));
                                break;
                            default:
                                break;
                        }

                        if (filter.logicalOperators[i] === 1) {
                            columnOperator = ' OR ';
                        }
                    }

                    whereParts.push('(' + columnWhereParts.join(columnOperator) + ')');
                }

                if (whereParts.length > 0) {
                    let joinWith = ' AND ';

                    if (details.filterOperator && details.filterOperator.toLowerCase() === 'or') {
                        joinWith = ' OR ';
                    }

                    sqlQuery['where'] = ' WHERE ' + whereParts.join(joinWith);
                }
                else {
                    sqlQuery['where'] = '';
                }
            }
            else {
                sqlQuery['where'] = '';
            }

            if (details.action === 'dataBind' && that.virtualDataSourceOnExpand && that.keyDataField && that.parentDataField) {
                if (!sqlQuery['where']) {
                    sqlQuery['where'] = ' WHERE ' + that.parentDataField + ' IS NULL';
                }
                else {
                    sqlQuery['where'] += ' AND ' + that.parentDataField + ' IS NULL';
                }
            }

            if (details.action === 'update') {
                const setParts = [];

                for (let dataField in details.edit.row) {
                    if (dataField === '$' || dataField === that.id) {
                        continue;
                    }

                    setParts.push(dataField + '=\'' + details.edit.row[dataField] + '\'');
                }

                sqlQuery['update'] = ' SET ' + setParts.join(', ') + ' WHERE ' + that.id + '=' + details.edit.row[that.id];
            }

            if (details.action === 'remove') {
                const whereParts = [];

                for (let dataField in details.edit.row) {
                    if (dataField === '$' || dataField === that.id) {
                        continue;
                    }

                    whereParts.push(dataField + ' = \'' + details.edit.row[dataField] + '\'');
                }

                sqlQuery['remove'] = ' WHERE ' + that.id + '=' + details.edit.row[that.id];
            }

            if (details.action === 'add') {
                const values = [];
                const columns = [];

                for (let dataField in details.edit.row) {
                    if (dataField === '$') {
                        continue;
                    }

                    columns.push(dataField);
                    values.push('\'' + details.edit.row[dataField] + '\'');
                }

                sqlQuery['add'] = ' (' + columns.join(', ') + ') ' + ' VALUES (' + values.join(', ') + ')';
            }

            if (details.action === 'expand') {
                if (that.groupDataField) {
                    if (!sqlQuery['where']) {
                        sqlQuery['where'] = ' WHERE ' + that.groupDataField + ' = \'' + row[that.groupDataField] + '\'';
                    }
                    else {
                        sqlQuery['where'] += ' AND ' + that.groupDataField + ' = \'' + row[that.groupDataField] + '\'';
                    }
                }
                else if (that.keyDataField && that.parentDataField) {
                    if (!sqlQuery['where']) {
                        sqlQuery['where'] = ' WHERE ' + that.parentDataField + ' = ' + row[that.keyDataField];
                    }
                    else {
                        sqlQuery['where'] += ' AND ' + that.parentDataField + ' = ' + row[that.keyDataField];
                    }
                }
                else {
                    if (!sqlQuery['where']) {
                        sqlQuery['where'] = ' WHERE ' + row.groupDataField + ' = \'' + row.label + '\'';
                    }
                    else {
                        sqlQuery['where'] += ' AND ' + row.groupDataField + ' = \'' + row.label + '\'';
                    }
                }

                sqlQuery['limit'] = '';
                sqlQuery['groupBy'] = '';

                that.virtualDataSourceOnExpand(getDataSource, {
                    first: first,
                    last: last,
                    row: details.row,
                    sorting: details.sorting,
                    filtering: details.filtering,
                    grouping: details.grouping,
                    action: details.action,
                    query: sqlQuery
                });

            }
            else {
                that.virtualDataSource(getDataSource, {
                    first: first,
                    last: last,
                    sorting: details.sorting,
                    filtering: details.filtering,
                    filterOperator: details.filterOperator || 'and',
                    grouping: details.grouping,
                    action: details.action,
                    query: sqlQuery
                });
            }
        }
        else {
            requestCallback();
        }
    }

    add(item, parentId) {
        const that = this;

        if (!item) {
            return;
        }

        let result = true;

        const addItem = function (item) {
            const itemObject = that._getDataItem(item, that.boundSource.length);

            if (that.dataItemById[itemObject.$.id]) {
                if (parentId !== undefined) {
                    itemObject.$.parentId = parentId;
                }
                return null;
            }

            that[that.boundSource.length] = itemObject;
            that.dataItemById[itemObject.$.id] = itemObject;

            const pushResult = that.boundSource.push(itemObject);

            if (parentId !== undefined) {
                itemObject.$.parentId = parentId;
            }

            if (!pushResult) {
                result = false;
            }

            return itemObject;
        }

        if (item.length) {
            let itemObjects = [];

            for (let i = 0; i < item.length; i++) {
                const itemObject = addItem(item[i]);

                if (itemObject) {
                    itemObjects.push(itemObject);
                }
            }

            that._notify({ action: 'add', data: itemObjects });
        }
        else {
            const itemObject = addItem(item);

            that._notify({ action: 'add', data: itemObject });
        }

        that.refreshHierarchy();

        return result;
    }

    refreshIndexes() {
        const that = this;

        for (let i = 0; i < that.boundSource.length; i++) {
            that[i] = that.boundSource[i];
            that[i].$.index = i;
            that.dataItemById[that[i].$.id] = that[i];
        }

        let i = that.boundSource.length;

        while (that[i]) {
            delete that[i];
            i++;
        }
    }

    removeLast() {
        const that = this;

        delete that[that.boundSource.length - 1];
        const result = that.boundSource.pop();
        delete that.dataItemById[result.$.id];

        that._notify({ action: 'removeLast', data: result });

        that.refreshHierarchy();

        return result;
    }

    remove(item) {
        const that = this;
        const index = that.indexOf(item);

        that.removeAt(index);
    }

    removeAt(index) {
        const that = this;

        const item = that.boundSource[index];

        if (!item) {
            throw new Error('Invalid Item Index');
        }

        that.boundSource.splice(index, 1);
        delete that.dataItemById[item.$.id];
        that.refreshIndexes();

        that._notify({ action: 'remove', index: index, data: item });

        that.refreshHierarchy();
    }

    update(index, dataSourceItem) {
        const that = this;

        if (Smart.Utilities.Types.isArray(index) && Smart.Utilities.Types.isArray(dataSourceItem)) {
            if (index.length === 0 && dataSourceItem.length === 0) {
                that.refreshHierarchy();
                return;
            }
        }

        if (dataSourceItem.length && index.length) {
            let itemObjects = [];

            for (let i = 0; i < index.length; i++) {
                const itemObject = that._getDataItem(dataSourceItem[i], index[i]);
                const currentIndex = index[i];

                itemObjects.push(itemObject);

                that.boundSource[currentIndex] = itemObject;
                that[currentIndex] = that.boundSource[currentIndex];
                that.dataItemById[itemObject.$.id] = that[currentIndex];
            }

            that._notify({ action: 'update', index: index, data: itemObjects });

            that.refreshHierarchy();

            return;
        }

        const itemObject = that._getDataItem(dataSourceItem, index);

        const id = that.boundSource[index].$.id;

        itemObject.$.id = id;
        that.boundSource[index] = itemObject;
        that[index] = that.boundSource[index];
        that.dataItemById[itemObject.$.id] = that[index];

        that._notify({ action: 'update', index: index, data: itemObject });

        that.refreshHierarchy();

        return itemObject;
    }

    insert(index, item) {
        const that = this;

        item = that._getDataItem(item, index);

        const result = that.boundSource.splice(index, 0, item);

        that.refreshIndexes();

        that._notify({ action: 'insert', index: index, data: item });

        that.refreshHierarchy();

        return result;
    }

    move(from, to) {
        if (to > from && to - from === 1 || from === to) {
            return;
        }

        const that = this,
            recordToMove = that.boundSource.splice(from, 1)[0];

        if (to > from) {
            to--;
            that.boundSource.splice(to, 0, recordToMove);
        }
        else {
            that.boundSource.splice(to, 0, recordToMove);
        }

        that.refreshIndexes();

        that._notify({ action: 'move', index: to, data: that.boundSource[to] });

        that.refreshHierarchy();
    }

    indexOf(item) {
        const that = this;
        const index = that.boundSource.indexOf(item);

        return index;
    }

    get length() {
        const that = this;

        if (that.virtualDataSourceLength !== undefined) {
            return that.virtualDataSourceLength;
        }

        if (that.dataSourceLength) {
            return that.dataSourceLength;
        }

        if (typeof (that.dataSource) === 'number') {
            return that.dataSource;
        }

        if (that.bindingCompleted) {
            return that.boundSource.length;
        }

        if (that.dataSource && typeof that.dataSource !== 'string' && that.dataSource.length) {
            return that.dataSource.length;
        }

        return that.boundSource.length;
    }

    clear() {
        const that = this;

        if (!that.isInitialized) {
            that._cachedValues = [];
            that.dataItemById = [];
            return;
        }

        for (let i = 0; i < that.boundSource.length; i++) {
            delete that[i];
        }

        that._expandedRowIds = [];
        that._cachedValues = [];
        that.boundSource = that.observable ? new Smart.ObservableArray() : [];
        that.dataItemById = [];
        that.refreshHierarchy();
    }

    _getId(id, item, index) {
        if (id !== null && id.name !== undefined) {
            if (id.name && item.getAttribute) {
                let result = item.getAttribute(id.name);
                if (result !== null && result.toString().length > 0) {
                    return result;
                }
                else if (id.map) {
                    try {
                        let result = item.getAttribute(id.map);
                        if (result !== null && result.toString().length > 0) {
                            return result;
                        }
                    }
                    catch (error) {
                        return index;
                    }
                }
                return;
            }
        }

        if (id) {
            if (id.toString().length > 0 && item.getAttribute) {
                let result = item.getAttribute(id);
                if (result !== null && result.toString().length > 0) {
                    return result.trim().split(' ').join('').replace(/([ #;?%&,.+*~\':'!^$[\]()=>|\/@])/g, '');
                }
                else {
                    let splitMap = id.split(this.mapChar);
                    if (splitMap.length > 1) {
                        let d = item;
                        for (let p = 0; p < splitMap.length; p++) {
                            if (d !== undefined) {
                                d = d[splitMap[p]];
                            }
                        }
                        if (d !== undefined) {
                            return d;
                        }
                    }
                    else {
                        if (item[id] !== undefined) {
                            return item[id];
                        }
                    }
                }
            }
        }

        return index;
    }

    _buildHierarchy() {
        const that = this;

        if (!that.reservedNames) {
            that.reservedNames = {
                leaf: 'leaf',
                parent: 'parent',
                expanded: 'expanded',
                checked: 'checked',
                selected: 'selected',
                level: 'level',
                icon: 'icon',
                data: 'data'
            }
        }
        else {
            const names = that.reservedNames;

            if (!names.leaf) {
                names.leaf = 'leaf';
            }
            if (!names.parent) {
                names.parent = 'parent';
            }
            if (!names.expanded) {
                names.expanded = 'expanded';
            }
            if (!names.checked) {
                names.checked = 'checked';
            }
            if (!names.selected) {
                names.selected = 'selected';
            }
            if (!names.level) {
                names.level = 'level';
            }
            if (!names.data) {
                names.data = 'data';
            }

        }

        const names = that.reservedNames;

        if (that.childrenDataField) {
            const hierarchy = [];

            for (let i = 0; i < that.boundSource.length; i++) {
                const item = Object.assign({}, that.boundSource[i]);

                if (!item) {
                    continue;
                }

                hierarchy.push(item);

                const addItems = function (item) {
                    const splitMap = that.childrenDataField.split(that.mapChar);
                    let children = null;

                    if (splitMap.length > 1) {
                        let data = item;

                        for (let p = 0; p < splitMap.length; p++) {
                            if (data !== undefined) {
                                data = data[splitMap[p]];
                            }
                        }

                        children = data;
                    }
                    else {
                        children = item['children'];
                    }

                    item['children'] = children;

                    if (item['children'] === null || item['children'] === undefined || (item['children'] && item['children'].length === 0)) {
                        item[names.leaf] = true;
                    }
                }

                addItems(item);
                item[names.level] = 0;

                if (!item.$) {
                    item.$ = {};
                }

                item[names.parent] = null;
                item[names.data] = item;

                if (item[names.expanded] === undefined) {
                    item[names.expanded] = false;

                    if (that.expandHierarchy) {
                        item[names.expanded] = true;
                    }
                }

                const drillThrough = function (parent, children) {
                    if (!children) {
                        parent['children'] = new Array();
                        return;
                    }

                    for (let i = 0; i < children.length; i++) {
                        let item = that._getDataItem(children[i], i);

                        if (!item) {
                            continue;
                        }

                        addItems(item);

                        if (!that.id) {
                            const generateKey = function () {
                                const S4 = function () {
                                    return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
                                };
                                return S4() + S4();
                            };

                            item.$.id = generateKey();
                        }
                        item[names.level] = parent[names.level] + 1;
                        item[names.parent] = parent;
                        item[names.data] = Object.assign({}, item);

                        if (parent) {
                            parent['children'][i] = item;
                        }


                        if (item[names.expanded] === undefined) {
                            item[names.expanded] = false;
                            if (that.expandHierarchy) {
                                item[names.expanded] = true;
                            }
                        }


                        drillThrough(item, item['children']);
                    }
                }

                drillThrough(item, item['children']);
            }


            that.boundHierarchy = hierarchy;

            if (!that._boundSourceUpdate) {
                for (let i = 0; i < that.boundHierarchy.length; i++) {
                    const item = that.boundHierarchy[i];

                    if (item.children) {
                        const drillThrough = function (item) {
                            if (!that.dataItemById[item.$.id]) {
                                that.boundSource.canNotify = false;
                                that.dataItemById[item.$.id] = item;
                                //     that[ that.boundSource.length ] = item;
                                //     that.boundSource.push( item );
                                that.boundSource.canNotify = true;
                            }

                            if (item.children) {
                                for (let i = 0; i < item.children.length; i++) {
                                    const child = item.children[i];

                                    if (child.children) {
                                        drillThrough(child);
                                    }
                                }
                            }
                        }

                        drillThrough(item);
                    }
                }

                that._boundSourceUpdate = true;
            }
        }

        if (that.xmlRoot && that.dataSourceType === 'xml') {
            that.boundHierarchy = this._getHierarchy('uid', '_parentuid', 'children', null, that.boundSource);
        }

        if (that.keyDataField && that.parentDataField) {
            ['keyDataField', 'parentDataField'].forEach((fieldName) => {
                switch (that[fieldName]) {
                    case 'parent':
                        {
                            that[fieldName] = '$' + that[fieldName];
                        }
                }
            });

            that.boundHierarchy = this._getHierarchy(that.keyDataField, that.parentDataField, 'children', null, that.boundSource);

            return;
        }

        if (that.groupBy && that.groupBy.length > 0) {
            that.boundHierarchy = this._getGroupHierarchy(that.groupBy, 'children', 'label', null, 'data', null, 'parent', that.boundSource);

            return;
        }

        if (that.virtualDataSourceOnExpand) {
            that.boundHierarchy = this._getHierarchy('id', 'parentId', 'children', null, that.boundSource);
        }
    }


    _getGroupHierarchy(groups, collectionName, groupName, mappingFields, itemName, valueName, parentName, data, startIndex) {
        let that = this;

        if (!startIndex) {
            startIndex = 0;
        }

        let names = that.reservedNames;

        const guid = function () {
            function s4() {
                return Math.floor((1 + Math.random()) * 0x10000)
                    .toString(16)
                    .substring(1);
            }

            return s4() + s4() + '-' + s4() + '-' + s4() + '-' + s4() + '-' + s4() + s4() + s4();
        }

        let groupHashCodes = new Array();
        for (let iGroupColumn = 0; iGroupColumn < groups.length; iGroupColumn++) {
            groupHashCodes[iGroupColumn] = guid();
        }

        if (!collectionName) {
            collectionName = 'children';
        }

        if (!groupName) {
            groupName = 'group';
        }

        if (!itemName) {
            itemName = 'item';
        }

        if (!parentName) {
            parentName = 'parent';
        }

        if (undefined === valueName) {
            valueName = 'value';
        }

        const groupboundSource = new Array();
        const hashItemGroups = new Array();

        let groupboundSourceIndex = 0;

        const getItem = function (item) {
            let itemObj = item;
            if (mappingFields) {
                for (let mappingField in mappingFields) {
                    const mappingObject = mappingFields[mappingField];

                    if (mappingObject.name && mappingObject.map) {
                        itemObj[mappingObject.map] = itemObj[mappingObject.name];
                    }
                }
            }

            return itemObj;
        }

        let hasNullValues = false;

        for (let obj = 0; obj < data.length; obj++) {
            let item = Object.assign({}, getItem(data[obj]));

            item[names.leaf] = false;

            let itemKeysHierarchy = new Array();
            let keys = 0;

            for (let iGroupColumn = 0; iGroupColumn < groups.length; iGroupColumn++) {
                const group = groups[iGroupColumn];
                const value = item[group];

                if (null === value) {
                    itemKeysHierarchy[keys++] = { value: null, group: group, hash: groupHashCodes[iGroupColumn] };

                    if (iGroupColumn === 0) {
                        hasNullValues = true;
                    }
                    continue;
                }

                if (value === '' || value === undefined && iGroupColumn === 0) {
                    hasNullValues = true;
                }

                itemKeysHierarchy[keys++] = { value: value, group: group, hash: groupHashCodes[iGroupColumn] };
            }

            if (itemKeysHierarchy.length !== groups.length) {
                break;
            }

            let parentItem = null;
            let lookupKey = '';

            for (let q = 0; q < itemKeysHierarchy.length; q++) {
                const itemKey = itemKeysHierarchy[q].value;
                const groupDataField = itemKeysHierarchy[q].group;
                const columnHash = itemKeysHierarchy[q].hash;

                lookupKey = lookupKey + '_' + columnHash + '_' + itemKey;

                if (hashItemGroups[lookupKey] !== undefined && hashItemGroups[lookupKey] !== null) {
                    parentItem = hashItemGroups[lookupKey];
                    continue;
                }

                if (parentItem === null) {
                    parentItem = { $: {} };

                    parentItem[names.level] = 0;
                    parentItem[names.leaf] = false;
                    parentItem[parentName] = null;
                    parentItem[groupName] = itemKey;
                    parentItem[itemName] = item;
                    parentItem['groupDataField'] = groupDataField;

                    if (!parentItem[groupDataField]) {
                        parentItem[groupDataField] = parentItem.data[groupDataField];
                    }

                    if (item[names.expanded] !== undefined) {
                        parentItem[names.expanded] = item[names.expanded];
                    }
                    else {
                        parentItem[names.expanded] = false;
                        if (that.expandHierarchy) {
                            parentItem[names.expanded] = true;
                        }
                    }

                    if (valueName) {
                        parentItem[valueName] = item[valueName];
                    }

                    parentItem[collectionName] = new Array();

                    let uid = groupboundSource.length + startIndex;

                    if (!this.id || typeof item.$.id === 'number' || isFinite(item.$.id)) {
                        uid = 'Item' + uid;
                    }
                    if (parentItem.$.id === undefined) {
                        parentItem.$.id = uid;
                    }

                    if (that._expandedRowIds[parentItem.$.id]) {
                        parentItem[names.expanded] = true;
                    }


                    groupboundSource[groupboundSourceIndex++] = parentItem;
                }
                else {
                    const subItem = { $: {} };

                    subItem[names.level] = parentItem[names.level] + 1;
                    subItem[parentName] = parentItem;
                    subItem[groupName] = itemKey;
                    subItem[collectionName] = new Array();
                    subItem[itemName] = item;
                    subItem['groupDataField'] = groupDataField;
                    subItem[names.leaf] = false;

                    if (!subItem[groupDataField]) {
                        subItem[groupDataField] = subItem.data[groupDataField];
                    }

                    if (item[names.expanded] !== undefined) {
                        subItem[names.expanded] = item[names.expanded];
                    }
                    else {
                        subItem[names.expanded] = false;

                        if (that.expandHierarchy) {
                            subItem[names.expanded] = true;
                        }
                    }

                    if (valueName) {
                        subItem[valueName] = item[valueName];
                    }

                    if (subItem.$.id === undefined) {
                        subItem.$.id = parentItem.$.id + '_' + parentItem[collectionName].length;
                    }

                    if (that._expandedRowIds[subItem.$.id]) {
                        subItem[names.expanded] = true;
                    }


                    parentItem[collectionName][parentItem[collectionName].length] = subItem;
                    parentItem = subItem;
                }

                hashItemGroups[lookupKey] = parentItem;
            }

            if (item) {
                item[names.leaf] = true;
            }

            if (parentItem !== null) {
                if (this.id === null) {
                    if (undefined === item.$.id) {
                        item.$.id = parentItem.$.id + '_' + parentItem[collectionName].length;
                    }
                }
                else {
                    if (undefined === item.$.id) {
                        if (item.$.id.toString().indexOf(parentItem.$.id) === -1) {
                            item.$.id = parentItem.$.id + '_' + item.$.id;
                        }
                    }
                }

                item[parentName] = parentItem;
                item[names.level] = parentItem[names.level] + 1;
                parentItem[collectionName][parentItem[collectionName].length] = item;
            }
            else {
                if (undefined === item.$.id) {
                    item.$.id = guid();
                }
            }
        }

        if (hasNullValues) {
            let emptyItem = null;
            let updateOrderId = false;
            for (let i = 0; i < groupboundSource.length; i++) {
                const groupItem = groupboundSource[i];

                if (!groupItem[groupItem.groupDataField]) {
                    emptyItem = groupItem;

                    if (emptyItem.$.id.toString().startsWith('Item')) {
                        updateOrderId = true;
                    }

                    groupboundSource.splice(i, 1);
                    break;
                }
            }

            if (emptyItem) {
                groupboundSource.push(emptyItem);

                if (updateOrderId) {
                    for (let i = 0; i < groupboundSource.length; i++) {
                        const groupItem = groupboundSource[i];

                        groupItem.$.id = 'Item' + i;
                    }
                }
            }
        }

        return groupboundSource;
    }

    _getHierarchy(fieldName, parentFieldName, collectionName, mappingFields, boundSource) {
        const that = this;

        const databoundHierarchy = new Array();
        let flatData = this.boundSource;

        if (boundSource) {
            flatData = boundSource;
        }

        if (this.boundSource.length === 0)
            return null;

        const childrenName = collectionName !== null ? collectionName : 'children';

        let items = new Array();
        let data = flatData;
        let dataLength = data.length;
        let names = that.reservedNames;

        const getItem = function (item) {
            let itemObj = item;
            if (mappingFields) {
                for (let mappingField in mappingFields) {
                    const mappingObject = mappingFields[mappingField];

                    if (mappingObject.name && mappingObject.map) {
                        itemObj[mappingObject.map] = itemObj[mappingObject.name];
                    }
                }
            }

            return itemObj;
        }

        // build hierarchical source.
        for (let i = 0; i < dataLength; i++) {
            let item = data[i];
            let parentId = item[parentFieldName];
            let id = item[fieldName];

            if (parentFieldName === 'parentId') {
                parentId = item.$.parentId;
            }

            if (fieldName === 'id') {
                id = item.$.id;
            }

            item[childrenName] = new Array();

            if (that.expandHierarchy && item[names.expanded] === undefined) {
                item[names.expanded] = that.expandHierarchy;
            }

            items[id] = { parentId: parentId, item: item };
        }

        for (let i = 0; i < dataLength; i++) {
            const item = data[i];
            let parentId = item[parentFieldName];
            let id = item[fieldName];

            if (parentFieldName === 'parentId') {
                parentId = item.$.parentId;
            }

            if (fieldName === 'id') {
                id = item.$.id;
            }

            if (items[parentId] !== undefined) {
                let item = { parentId: parentId, item: items[id].item };
                let parentItem = items[parentId].item;
                if (!parentItem[childrenName]) {
                    parentItem[childrenName] = new Array();
                }
                let length = parentItem[childrenName].length;
                item = item.item;

                if (!names) {
                    item.parent = parentItem;
                }
                else {
                    item[names.parent] = parentItem;
                }

                const itemObj = getItem(item);

                parentItem[childrenName][length] = itemObj;
                items[parentId].item = parentItem;
                items[id].item = item;

            }
            else {
                let item = items[id].item;
                if (!names) {
                    if (item.parent === undefined) {
                        item.parent = null;
                    }
                }
                else {
                    if (item[names.parent] === undefined) {
                        item[names.parent] = null;
                    }
                }

                const itemObj = getItem(item);

                if (!names) {
                    itemObj.level = 0;
                }
                else {
                    itemObj[names.level] = 0;
                }

                databoundHierarchy[databoundHierarchy.length] = itemObj;
            }
        }
        if (databoundHierarchy.length !== 0) {
            let updateLevels = function (level, children) {
                for (let i = 0; i < children.length; i++) {
                    const child = children[i];

                    if (!names) {
                        child.level = level;
                    }
                    else {
                        child[names.level] = level;
                    }

                    const childChildren = child[childrenName];

                    if (childChildren) {
                        if (childChildren.length > 0) {
                            child.leaf = false;
                            updateLevels(level + 1, childChildren);
                        }
                        else {
                            if (that.virtualDataSourceOnExpand) {
                                if (child.leaf === undefined) {
                                    child.leaf = false;
                                }
                            }
                            else {
                                if (!names) {
                                    child.leaf = true;
                                }
                                else {
                                    child[names.leaf] = true;
                                }
                            }
                        }
                    }
                    else {
                        if (that.virtualDataSourceOnExpand) {
                            if (child.leaf === undefined) {
                                child.leaf = false;
                            }
                        }
                        else {
                            if (!names) {
                                child.leaf = true;
                            }
                            else {
                                child[names.leaf] = true;
                            }
                        }
                    }
                }
            };
            updateLevels(0, databoundHierarchy);
        }
        return databoundHierarchy;
    }

    summarize(summaryItems, boundSource) {
        const that = this;

        if (!Array.isArray(summaryItems)) {
            summaryItems = [summaryItems];
        }

        let tempSummaryItems = [];

        for (let i = 0; i < summaryItems.length; i++) {
            const summaryItem = summaryItems[i];

            for (let name in summaryItem) {
                const functions = summaryItem[name];

                tempSummaryItems.push({ dataField: name, functions: functions })
            }
        }

        summaryItems = tempSummaryItems;

        let data = {};
        let summaryByDataField = new Array();

        if (!boundSource) {
            boundSource = that.boundSource;
        }

        let length = boundSource.length;

        if (length === 0) {
            return;
        }

        if (length === undefined) {
            return;
        }

        for (let i = 0; i < length; i++) {
            let dataItem = boundSource[i];

            for (let j = 0; j < summaryItems.length; j++) {
                const summaryItem = summaryItems[j];
                let value = dataItem[summaryItem.dataField];

                if (summaryItem.functions) {
                    data[summaryItem.dataField] = data[summaryItem.dataField] || {};
                    summaryByDataField[summaryItem.dataField] = summaryByDataField[summaryItem.dataField] || 0;
                    summaryByDataField[summaryItem.dataField]++;

                    const _summaryItemFunction = function (summaryItemObject) {
                        for (let name in summaryItemObject) {
                            let oldValue = data[summaryItem.dataField][name];

                            if (oldValue === null || oldValue === undefined) {
                                data[summaryItem.dataField][name] = 0;
                                oldValue = 0;
                            }

                            if (typeof summaryItemObject[name] === 'function') {
                                oldValue = summaryItemObject[name](oldValue, value, summaryItem.dataField, dataItem);
                            }
                            data[summaryItem.dataField][name] = oldValue;
                        }
                    }

                    let canParse = parseFloat(value);

                    if (isNaN(canParse)) {
                        canParse = false;
                    }
                    else {
                        canParse = true;
                    }

                    if (canParse) {
                        value = parseFloat(value);
                    }

                    if (typeof value === 'number' && isFinite(value)) {
                        summaryItem.functions.forEach(function (summaryItemFunction) {
                            let oldValue = data[summaryItem.dataField][summaryItemFunction];

                            if (oldValue === null || oldValue === undefined) {
                                oldValue = 0;

                                if (summaryItemFunction === 'min') {
                                    oldValue = 9999999999999;
                                }

                                if (summaryItemFunction === 'max') {
                                    oldValue = -9999999999999;
                                }

                                if (summaryItemFunction === 'median') {
                                    oldValue = [];
                                }
                            }

                            if (summaryItemFunction === 'sum' || summaryItemFunction === 'avg' || summaryItemFunction === 'stdev'
                                || summaryItemFunction === 'stdevp' || summaryItemFunction === 'var' || summaryItemFunction === 'varp') {
                                oldValue += parseFloat(value);
                            }
                            else if (summaryItemFunction === 'product') {
                                if (i === 0)
                                    oldValue = parseFloat(value);
                                else
                                    oldValue *= parseFloat(value);
                            }
                            else if (summaryItemFunction === 'min') {
                                oldValue = Math.min(oldValue, parseFloat(value));
                            }
                            else if (summaryItemFunction === 'max') {
                                oldValue = Math.max(oldValue, parseFloat(value));
                            }
                            else if (summaryItemFunction === 'count') {
                                oldValue++;
                            }
                            else if (summaryItemFunction === 'median') {
                                oldValue.push(parseFloat(value));
                            }
                            else if (typeof (summaryItemFunction) === 'object') {
                                _summaryItemFunction(summaryItemFunction);
                                return;
                            }

                            data[summaryItem.dataField][summaryItemFunction] = oldValue;
                        });
                    }
                    else {
                        summaryItem.functions.forEach(function (summaryItemFunction) {
                            if (summaryItemFunction === 'min' || summaryItemFunction === 'max' || summaryItemFunction === 'count' || summaryItemFunction === 'product' || summaryItemFunction === 'sum'
                                || summaryItemFunction === 'avg' || summaryItemFunction === 'stdev'
                                || summaryItemFunction === 'stdevp' || summaryItemFunction === 'var' || summaryItemFunction === 'varp') {
                                if (value === null) {
                                    return true;
                                }

                                let oldValue = data[summaryItem.dataField][summaryItemFunction];

                                if (oldValue === null || oldValue === undefined) {
                                    oldValue = 0;
                                }

                                data[summaryItem.dataField][summaryItemFunction] = oldValue;

                                return true;
                            }

                            if (typeof (summaryItemFunction) === 'object') {
                                _summaryItemFunction(summaryItemFunction);
                            }
                        });
                    }
                }
            }
        }

        for (let j = 0; j < summaryItems.length; j++) {
            const summaryItem = summaryItems[j];

            if (!summaryItem.functions) {
                continue;
            }
            if (!data[summaryItem.dataField]) {
                data[summaryItem.dataField] = {};

                summaryItem.functions.forEach(function (summaryItemFunction) {
                    data[summaryItem.dataField][summaryItemFunction] = 0;
                });
            }

            if (data[summaryItem.dataField]['avg'] !== undefined) {
                const value = data[summaryItem.dataField]['avg'];
                const dataValues = summaryByDataField[summaryItem.dataField];

                if (dataValues === 0 || dataValues === undefined) {
                    data[summaryItem.dataField]['avg'] = 0;
                }
                else {
                    data[summaryItem.dataField]['avg'] = value / dataValues;
                }
            }
            else if (data[summaryItem.dataField]['count'] !== undefined) {
                data[summaryItem.dataField]['count'] = length;
            }
            else if (data[summaryItem.dataField]['median'] !== undefined) {
                let population = data[summaryItem.dataField]['median'];

                population.sort(function (a, b) {
                    return a - b;
                });

                data[summaryItem.dataField]['median'] =
                    0.5 * (population[Math.floor((population.length + 1) / 2) - 1] + population[Math.ceil((population.length + 1) / 2) - 1]);
            }

            // stdev, stdevp, var, varp.
            // stdev - Standard deviation on a sample.
            // varp - Variance on an entire population.
            // let - Variance on a sample.
            if (data[summaryItem.dataField]['stdev'] || data[summaryItem.dataField]['stdevp']
                || data[summaryItem.dataField]['var'] || data[summaryItem.dataField]['varp']) {
                summaryItem.functions.forEach(function (summaryItemFunction) {
                    if (summaryItemFunction === 'stdev' || summaryItemFunction === 'var' || summaryItemFunction === 'varp' || summaryItemFunction === 'stdevp') {
                        const value = data[summaryItem.dataField][summaryItemFunction];
                        const count = length;
                        const average = (value / length);
                        let sumSq = 0.0;

                        for (let i = 0; i < length; i++) {
                            let dataItem = boundSource[i];
                            let value = dataItem[summaryItem.dataField];

                            sumSq += (value - average) * (value - average);
                        }

                        let denominator = (summaryItemFunction === 'stdevp' || summaryItemFunction === 'varp') ? count : count - 1;

                        if (denominator === 0) {
                            denominator = 1;
                        }

                        if (summaryItemFunction === 'var' || summaryItemFunction === 'varp') {
                            data[summaryItem.dataField][summaryItemFunction] = sumSq / denominator;
                        }
                        else if (summaryItemFunction === 'stdevp' || summaryItemFunction === 'stdev') {
                            data[summaryItem.dataField][summaryItemFunction] = Math.sqrt(sumSq / denominator);
                        }
                    }
                });
            }
        }
        return data;
    }
    _getDataItem(dataSourceItem, index) {
        const that = this;
        const itemObject = {};
        const unboundMode = typeof (that.dataSource) === 'number' || that.dataSourceLength;

        if (!dataSourceItem) {
            return { $: { id: index, isEmpty: true, index: index } }
        }

        if (typeof dataSourceItem === 'string') {
            dataSourceItem = { '': dataSourceItem };
        }

        if (unboundMode) {
            for (let j = 0; j < that.dataFields.length; j++) {
                const dataField = that.dataFields ? that.dataFields[j] : {};

                itemObject[dataField.name] = '';
            }

            itemObject.$ = {};
            itemObject.$.id = index;
            itemObject.$.index = index;

            return itemObject;
        }

        const dataItem = dataSourceItem;

        if (dataItem.expanded !== undefined) {
            itemObject.expanded = dataItem.expanded;

            if (dataItem.expanded === 'true' || dataItem.expanded === true || dataItem.expanded === 1) {
                itemObject.expanded = true;
            }
            else {
                itemObject.expanded = false;
            }
        }

        if (that.childrenDataField) {
            if (dataItem[that.childrenDataField] !== undefined) {
                itemObject.children = dataItem[that.childrenDataField];
            }
        }
        else {
            if (dataItem.children !== undefined) {
                itemObject.children = dataItem.children;
            }
            else if (dataItem.items !== undefined) {
                itemObject.children = dataItem.items;
            }
        }
        if (dataItem.leaf !== undefined) {
            itemObject.leaf = dataItem.leaf;
        }

        if (dataItem.level !== undefined) {
            itemObject.level = dataItem.level;
        }

        if (that.keyDataField) {
            if (dataItem[that.keyDataField] !== undefined) {
                itemObject[that.keyDataField] = dataItem[that.keyDataField];
            }
        }

        if (that.parentDataField) {
            if (dataItem[that.parentDataField] !== undefined) {
                itemObject[that.parentDataField] = dataItem[that.parentDataField];
            }
        }

        if (that.dataFields.length === 0) {
            const names = Object.getOwnPropertyNames(dataSourceItem);

            for (let i = 0; i < names.length; i++) {
                if (names[i] === '$') {
                    continue;
                }

                that.dataFields.push({ name: names[i], dataType: 'string' });
            }
        }

        for (let j = 0; j < that.dataFields.length; j++) {
            const dataField = that.dataFields ? that.dataFields[j] : {};
            let value = '';

            if (undefined === dataField || dataField === null) {
                continue;
            }

            if (dataSourceItem.length) {
                value = dataSourceItem[j];
            }

            if (dataField.map) {
                let splitMap = dataField.map.split(that.mapChar);

                if (splitMap.length > 0) {
                    let dataMappedItem = dataItem;

                    for (let p = 0; p < splitMap.length; p++) {
                        if (!dataItem) {
                            continue;
                        }

                        dataMappedItem = dataMappedItem[splitMap[p]];
                    }

                    value = dataMappedItem;
                }
                else {
                    value = dataItem[dataField.map];
                }
            }

            if (value !== undefined && value !== null) {
                value = value.toString();
            }
            else {
                if (value === undefined && value !== null) {
                    value = '';
                }
            }


            let isEmptyString = false;
            // searches by both selectors when necessary.
            if (value === '') {
                isEmptyString = true;
                value = dataSourceItem[dataField.name];

                if (dataField.dataType === undefined) {
                    dataField.dataType = 'any';
                }

                if (value !== undefined && value !== null) {
                    if (dataField.dataType !== 'array') {
                        if (dataField.dataType === 'any') {
                            if (Array.isArray(dataSourceItem[dataField.name])) {
                                value = [...dataSourceItem[dataField.name]];
                            }
                            else if (typeof dataSourceItem[dataField.name] === 'object') {
                                value = Object.assign({}, dataSourceItem[dataField.name]);
                            }
                            else {
                                value = dataSourceItem[dataField.name];
                            }
                        }
                        else if (dataField.dataType !== 'date') {
                            value = value.toString();
                        }
                    }
                }
                else if (value === undefined && dataField.dataType === 'any') {
                    value = null;
                }
                else {
                    value = '';
                }
            }

            if (value === '[object Object]' && dataField.map && isEmptyString) {
                value = '';
            }

            if (that._cachedValues['' + value + '_' + dataField.dataType]) {
                value = that._cachedValues['' + value + '_' + dataField.dataType];
            }
            else {
                if (dataField.dataType === 'bool' || dataField.dataType === 'boolean') {
                    if (value === 'true' || value === '1') {
                        value = true;
                    }
                    else if (value === 'false' || value === '0') {
                        value = false;
                    }
                }
                else if (dataField.dataType === 'number') {
                    if (value.indexOf('.') === -1 && value.indexOf(',') >= 0) {
                        if (value.substring(1 + value.indexOf(',')).length < 3) {
                            value = value.replace(',', '.');
                        }
                    }
                    value = that.$document.deserialize('' + value, dataField.dataType, true);

                }
                else if (dataField.dataType !== 'any') {
                    value = that.$document.deserialize('' + value, dataField.dataType, true);
                }

                if (dataField.dataType !== 'any') {
                    that._cachedValues[value + '_' + dataField.dataType] = value;
                }
            }

            if (dataField.dataType !== 'string' && dataField.dataType !== 'boolean' && dataField.dataType !== 'bool' && dataField.dataType !== 'any') {
                if (isNaN(value) || value === -Infinity || value === Infinity) {
                    value = 0;
                }
            }

            if (dataField.name === 'parent' || dataField.name === 'group' || dataField.name === 'leaf' || dataField.name === 'level') {
                itemObject.$[dataField.name] = value;
                continue;
            }

            if (dataField.dataType === 'string') {
                if (/<.+?>/.test(value)) {
                    // sanitizing value when it contains HTML tags
                    if (value.replace) {
                        if (that.sanitizeHTML === 'all') {
                            value = value.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').
                                replace(/'/g, '&#39;').replace(/"/g, '&quot;');
                        }
                        else if (that.sanitizeHTML === 'blackList') {
                            value = window.Smart.Utilities.Core.sanitizeHTML(value);
                        }
                    }
                }
            }

            itemObject[dataField.name] = value;
        }

        let itemObjectId = index;

        if (that.id) {
            itemObjectId = dataItem[that.id];
            if (typeof (itemObjectId) === 'object') {
                itemObjectId = index;
            }

            if (dataItem.$ && dataItem.$.id) {
                itemObjectId = dataItem.$.id;
            }
        }
        else if (!that.virtualDataSource && that.dataItemById && that.dataItemById[itemObjectId]) {
            itemObjectId = that.length;

            while (that.dataItemById[itemObjectId]) {
                itemObjectId++;
            }
        }

        if (!itemObject.$) {
            itemObject.$ = {};
        }

        itemObject.$.id = itemObjectId;
        itemObject.$.index = index;

        return new Object(itemObject);
    }

    _bindToArray() {
        const that = this;

        const unboundMode = typeof (that.dataSource) === 'number' || that.dataSourceLength;
        const dataArray = [];

        that.boundSource.canNotify = false;

        const length = unboundMode ? (typeof that.dataSource === 'number' ? that.dataSource : that.dataSource.length) : (!that.observable ? that.dataSource.length : that.length);

        for (let i = 0; i < length; i++) {
            const dataSourceItem = unboundMode ? {} : that.dataSource[i];
            const itemObject = that._getDataItem(dataSourceItem, i);

            dataArray.push(itemObject);
        }

        if (unboundMode && that.dataSourceLength && that.dataSource.length > 0) {
            for (let i = 0; i < that.dataSource.length; i++) {
                const cell = that.dataSource[i].cell;
                const value = that.dataSource[i].value;

                const row = cell.replace(/[^0-9]/g, '');
                const dataField = cell.replace(/[0-9]/g, '');

                dataArray[row - 1][dataField] = value;
            }
        }

        that.boundSource = false === that.observable || that.virtualDataSource ? [] : new Smart.ObservableArray(dataArray);

        if (that.observable) {
            that.boundSource = new Smart.ObservableArray(dataArray);
        }

        for (let i = 0; i < length; i++) {
            if (!that.observable) {
                that[i] = dataArray[i];
            }
            else {
                that[i] = that.boundSource[i];
            }

            that.dataItemById[that[i].$.id] = that[i];
        }

        if (false === that.observable) {
            that.boundSource = dataArray;
        }

        that.boundSource.canNotify = true;
    }

    _bindToICS() {
        const that = this;
        const data = '' + that.dataSource;
        const rowDelimiter = that.rowDelimiter || '\n';
        const rowsWithNewRowDelimiter = data.split('\r');
        const records = [];

        let rows = data.split(rowDelimiter);
        let length = rows.length;

        if (length === 1 && rowsWithNewRowDelimiter.length > 1) {
            rows = rowsWithNewRowDelimiter;
            length = rows.length;
        }

        const stringToDate = function (until, ignoreTime) {
            var re = /^(\d{4})(\d{2})(\d{2})(T(\d{2})(\d{2})(\d{2})Z)?$/;
            var bits = re.exec(until);
            if (!bits) {
                throw new Error('Invalid UNTIL value: ' + until)
            }

            if (ignoreTime) {
                return new Date(bits[1], bits[2] - 1, bits[3], 0, 0, 0, 0);
            }

            return new Date(
                Date.UTC(bits[1],
                    bits[2] - 1,
                    bits[3],
                    bits[5] || 0,
                    bits[6] || 0,
                    bits[7] || 0
                ));
        }

        for (let i = 0; i < length; i++) {
            const record = rows[i].trim();
            if (record.indexOf('BEGIN:VEVENT') >= 0) {
                var appointment = {};
                continue;
            }

            if (!appointment) {
                continue;
            }

            if (record.indexOf('BEGIN:VALARM') >= 0) {
                var alarm = {};
                continue;
            }

            if (alarm) {
                if (record.indexOf('ACTION') >= 0) {
                    alarm.ACTION = record.substring(record.indexOf('ACTION') + 7);
                    continue;
                }
                else if (record.indexOf('DESCRIPTION') >= 0) {
                    alarm.DESCRIPTION = record.substring(record.indexOf('DESCRIPTION') + 12);
                    continue;
                }
                else if (record.indexOf('TRIGGER') >= 0) {
                    alarm.TRIGGER = record.substring(record.indexOf('TRIGGER') + 8);
                    continue;
                }

                if (record.indexOf('END:VALARM') >= 0) {
                    if (!appointment.ALARM) {
                        appointment.ALARM = [];
                    }

                    appointment.ALARM.push(alarm);
                    alarm = undefined;
                    continue;
                }
            }

            if (record.indexOf('SUMMARY') >= 0) {
                appointment.SUMMARY = record.substring(record.indexOf('SUMMARY') + 8);
                continue;
            }
            else if (record.indexOf('LOCATION') >= 0) {
                appointment.LOCATION = record.substring(record.indexOf('LOCATION') + 9);
                continue;
            }
            else if (record.indexOf('DESCRIPTION') >= 0) {
                appointment.DESCRIPTION = record.substring(record.indexOf('DESCRIPTION') + 12);
                continue;
            }
            else if (record.indexOf('RRULE') >= 0) {
                appointment.RRULE = record.substring(record.indexOf('RRULE') + 6);
                continue;
            }
            else if (record.indexOf('RECURRENCE-ID') >= 0) {
                appointment.RECURRENCEID = record.substring(record.indexOf('RECURRENCE-ID') + 14);
                continue;
            }
            else if (record.indexOf('RECURRENCE-ID;VALUE=DATE') >= 0) {
                appointment.RECURRENCEID = record.substring(record.indexOf('RECURRENCE-ID;VALUE=DATE') + 25);
                continue;
            }
            else if (record.indexOf('EXDATE;VALUE=DATE') >= 0) {
                const exDate = record.substring(record.indexOf('EXDATE;VALUE=DATE') + 20);
                appointment.EXDATE = exDate;
                continue;
            }
            else if (record.indexOf('EXDATE') >= 0) {
                const exDate = record.substring(record.indexOf('EXDATE') + 7);
                appointment.EXDATE = exDate;
                continue;
            }
            else if (record.indexOf('DTEND;VALUE=DATE') >= 0) {
                appointment.DTEND = stringToDate(record.substring(record.indexOf('DTEND;VALUE=DATE') + 17), true);
                continue;
            }
            else if (record.indexOf('DTEND') >= 0) {
                appointment.DTEND = stringToDate(record.substring(record.indexOf('DTEND') + 6));
                continue;
            }
            else if (record.indexOf('DTSTART;VALUE=DATE') >= 0) {
                appointment.DTSTART = stringToDate(record.substring(record.indexOf('DTSTART;VALUE=DATE') + 19), true);
                continue;
            }
            else if (record.indexOf('DTSTART') >= 0) {
                appointment.DTSTART = stringToDate(record.substring(record.indexOf('DTSTART') + 8));
                continue;
            }
            else if (record.indexOf('UID') >= 0) {
                appointment.uid = appointment.UID = record.substring(record.indexOf('UID') + 4);
                continue;
            }
            else if (record.indexOf('STATUS') >= 0) {
                appointment.STATUS = record.substring(record.indexOf('STATUS') + 7);
                continue;
            }
            else if (record.indexOf('TITLE') >= 0) {
                appointment.TITLE = record.substring(record.indexOf('TITLE') + 6);
                continue;
            }
            else if (record.indexOf('CATEGORIES') >= 0) {
                appointment.CATEGORIES = record.substring(record.indexOf('CATEGORIES') + 11);
                continue;
            }
            else if (record.indexOf('END:VEVENT') >= 0) {
                records.push(appointment);

                if (appointment.ALARM) {
                    appointment.ALARM = JSON.stringify(appointment.ALARM);
                }
                alarm = undefined;
                continue;
            }
        }

        that.canNotify = false;
        that.dataSource = records;
        that.canNotify = true;

        that._bindToArray();
    }

    _bindToJSON() {
        const that = this;

        const dataArray = [];

        const dataEntries = that.dataSource.length ? that.dataSource : Object.entries(that.dataSource);

        that.boundSource.canNotify = false;

        for (let i = 0; i < dataEntries.length; i++) {
            const dataSourceItem = dataEntries[i];
            const itemObject = that._getDataItem(dataSourceItem, i);

            dataArray.push(itemObject);
        }

        that.boundSource = false === that.observable || that.virtualDataSource ? [] : new Smart.ObservableArray(dataArray);

        for (let i = 0; i < that.length; i++) {
            that[i] = that.boundSource[i];
            that.dataItemById[that[i].$.id] = that[i];
        }

        if (false === that.observable) {
            that.boundSource = dataArray;
        }

        that.boundSource.canNotify = true;
    }

    sortBy(dataField, dataType, orderBy) {
        const that = this;

        if (!dataType) {
            for (let i = 0; i < that.dataFields.length; i++) {
                const field = that.dataFields[i];

                if (field.name === dataField) {
                    dataType = field.dataType;
                    break;
                }
            }
        }

        if (that.boundHierarchy) {
            if ((!dataField || dataField.length === 0) && that.groupBy.length > 0) {
                that.refreshHierarchy();
                return;
            }

            // refresh hierarchy, if summary rows are present.
            const flattenHierarchy = (hierarchy) => {
                let result = [];

                while (hierarchy) {
                    result = result.concat(hierarchy);

                    for (let i = 0; i < hierarchy.length; i++) {
                        if (hierarchy[i].children) {
                            result = result.concat(hierarchy[i].children);
                        }
                    }

                    hierarchy = hierarchy.children;
                }

                return result;
            }

            const flattenedHierarchy = flattenHierarchy(that.boundHierarchy);

            if (flattenedHierarchy) {
                const summaryItem = flattenedHierarchy.find((item) => {
                    if (item.summaryRow) {
                        return item;
                    }
                })

                if (summaryItem) {
                    that.refreshHierarchy();
                }
            }

            const sortBy = function (hierarchy) {
                that._sort(hierarchy, dataField, orderBy, dataType);

                for (let i = 0; i < hierarchy.length; i++) {
                    const item = hierarchy[i];

                    if (item['children']) {
                        sortBy(item['children'], dataField, orderBy, dataType);
                    }
                }
            }

            sortBy(that.boundHierarchy);
        }
        else {
            that._sort(that.boundSource, dataField, orderBy, dataType);
        }
    }

    _createFilter(dataType, filterExpressions) {
        const filterOperators = {
            '=': 'EQUAL',
            '<>': 'NOT_EQUAL',
            '<': 'LESS_THAN',
            '>': 'GREATER_THAN',
            '<=': 'LESS_THAN_OR_EQUAL',
            '>=': 'GREATER_THAN_OR_EQUAL',
            'equal': 'EQUAL',
            'not equal': 'NOT_EQUAL',
            'less than': 'LESS_THAN',
            'greater than': 'GREATER_THAN',
            'greater than or equal': 'GREATER_THAN_OR_EQUAL',
            'less than or equal': 'LESS_THAN_OR_EQUAL',
            'starts with': 'STARTS_WITH',
            'ends with': 'ENDS_WITH',
            'notEqual': 'NOT_EQUAL',
            'lessThan': 'LESS_THAN',
            'greaterThan': 'GREATER_THAN',
            'greaterThanOrEqual': 'GREATER_THAN_OR_EQUAL',
            'lessThanOrEqual': 'LESS_THAN_OR_EQUAL',
            'null': 'null',
            '': 'EMPTY',
            'isblank': 'EMPTY',
            'isnotblank': 'NOT_EMPTY',
            'isBlank': 'EMPTY',
            'isNotBlank': 'NOT_EMPTY',
            'contains': 'CONTAINS',
            'notcontains': 'DOES_NOT_CONTAIN',
            'notContains': 'DOES_NOT_CONTAIN',
            'startswith': 'STARTS_WITH',
            'endswith': 'ENDS_WITH',
            'startsWith': 'STARTS_WITH',
            'endsWith': 'ENDS_WITH',
            'NULL': 'NULL',
            'NOT_NULL': 'NOT_NULL'
        };

        let filterExpressionsArray = [];

        for (let i = 0; i < filterExpressions.length; i++) {
            const filterExpression = filterExpressions[i];

            const filterExpressionParts = filterExpression.indexOf('"') === -1 ? filterExpression.split(' ') : filterExpression.split('"');
            let filter = [];

            for (let j = 0; j < filterExpressionParts.length; j++) {
                const part = filterExpressionParts[j];

                if (part !== '') {
                    filter.push(part.trim());
                }
            }

            filterExpressionsArray.push(filter);
        }

        const filterGroup = new Smart.FilterGroup();
        const filterGroupOperators = [];
        const filterSubGroups = [];

        for (let i = 0; i < filterExpressionsArray.length; i++) {
            const filterExpression = filterExpressionsArray[i];


            if (filterExpression.length > 1) {
                const filterSubGroup = new Smart.FilterGroup();

                let operator = 'and';
                let filterExpressionPartsCounter = 0;

                for (let j = 0; j < filterExpression.length; j++) {
                    const value = filterExpression[j];

                    if (value === 'and' || value === 'or') {
                        operator = value;
                        continue;
                    }

                    filterExpressionPartsCounter++;

                    if (filterExpressionPartsCounter === 2) {
                        const filter = filterSubGroup.createFilter(dataType, value, filterOperators[filterExpression[j - 1]]);

                        filterExpressionPartsCounter = 0;

                        if (operator) {
                            filterSubGroup.addFilter(operator, filter);
                        }
                    }
                }

                filterSubGroups.push(filterSubGroup);
            }
            else {
                const filterGroupOperator = filterExpression[0];

                if (filterGroupOperator !== 'and' && filterGroupOperator !== 'or') {
                    throw new Error('Filter Exprresion expects "AND" or "OR", but the token is: ' + filterGroupOperator);
                }

                filterGroupOperators.push(filterGroupOperator);
            }
        }

        let operatorsCounter = 0;

        if (filterSubGroups.length === 1) {
            return filterSubGroups[0];
        }

        for (let i = 0; i < filterSubGroups.length; i++) {
            let operator = filterGroupOperators[operatorsCounter];

            if ((i + 1) % 2 === 0) {
                operatorsCounter++;
            }

            if (!operator) {
                operator = 'and';
            }

            filterGroup.addFilter(operator, filterSubGroups[i]);
        }

        return filterGroup;
    }

    filterBy(dataField, ...filterExpressions) {
        const that = this;


        const dataType = (() => {
            for (let i = 0; i < that.dataFields.length; i++) {
                const field = that.dataFields[i];

                if (field.name === dataField) {
                    return field.dataType;
                }
            }
        })();


        const filterGroup = that._createFilter(dataType, filterExpressions);

        let filteredData = that.boundSource.filter((value) => {
            const evaluation = filterGroup.evaluate(value[dataField]);

            return evaluation;
        });

        return filteredData;
    }

    _filter(filters, operator = 'and') {
        const that = this;
        const filterGroups = [];
        const dataFields = [];

        if (filters.length === 0) {
            that.clearFilter();
            return;
        }

        const dataType = (dataField) => {
            for (let i = 0; i < that.dataFields.length; i++) {
                const field = that.dataFields[i];

                if (field.name === dataField) {
                    return field.dataType;
                }
            }
        };
        let defaultResult, operatorSpecificEval;

        if (operator === 'and') {
            defaultResult = true;
            operatorSpecificEval = function (result, filterGroup, row) {
                return result && filterGroup.evaluate(row[filterGroup.dataField]);
            };
        }
        else {
            defaultResult = false;
            operatorSpecificEval = function (result, filterGroup, row) {
                return result || filterGroup.evaluate(row[filterGroup.dataField]);
            };
        }

        for (let i = 0; i < filters.length; i++) {
            const filter = filters[i];
            const dataField = filter[0];
            let filterGroup = null;

            if (filter[1] instanceof Smart.FilterGroup) {
                filterGroup = filter[1];
            }
            else {
                filterGroup = that._createFilter(dataType(dataField), filter.splice(1));
            }

            if (filterGroup) {
                dataFields.push(dataField);
                filterGroup.dataField = dataField;
                filterGroups.push(filterGroup);
            }
        }

        if (that.boundHierarchy) {
            const filter = function (row) {
                let result = defaultResult;

                for (let j = 0; j < filterGroups.length; j++) {
                    const filterGroup = filterGroups[j];

                    result = operatorSpecificEval(result, filterGroup, row);
                }

                row.$.filtered = result;

                return result;
            }

            const filterBy = function (hierarchy, parentItem, root) {
                let filteredCount = 0;

                for (let i = 0; i < hierarchy.length; i++) {
                    const item = hierarchy[i];

                    filter(item);

                    if (item.$.filtered) {
                        filteredCount++;
                    }

                    if (item['children']) {
                        filterBy(item['children'], item, parentItem);
                    }
                }

                if (filteredCount > 0 && that.groupBy.length > 0 && parentItem) {
                    parentItem.$.filtered = true;

                    if (root && !root.$.filtered) {
                        root.$.filtered = true;
                    }
                }
                else {
                    if (filteredCount > 0 && filteredCount !== hierarchy.length && parentItem) {
                        parentItem.$.filtered = null;

                        if (root && !root.$.filtered) {
                            root.$.filtered = null;
                        }
                    }
                    else if (filteredCount > 0 && filteredCount === hierarchy.length && parentItem) {
                        parentItem.$.filtered = true;

                        if (root && !root.$.filtered) {
                            root.$.filtered = true;
                        }
                    }
                }
            }

            filterBy(that.boundHierarchy, null, null);
        }
        else {
            for (let i = 0; i < that.boundSource.length; i++) {
                const row = that.boundSource[i];

                let result = defaultResult;

                for (let j = 0; j < filterGroups.length; j++) {
                    const filterGroup = filterGroups[j];

                    result = operatorSpecificEval(result, filterGroup, row);
                }

                row.$.filtered = result;
            }
        }

        if (that.onFilter) {
            that.onFilter()
        }
    }

    clearGroup() {
        const that = this;

        that.groupBy = [];
        that.boundHierarchy = null;
        that.refreshHierarchy();

        if (that.onGroup) {
            that.onGroup()
        }
    }

    clearFilter() {
        const that = this;

        for (let i = 0; i < that.boundSource.length; i++) {
            const row = that.boundSource[i];

            row.$.filtered = true;
        }

        if (that.boundHierarchy) {
            const filterBy = function (hierarchy, parentItem, root) {
                //let filteredCount = 0;

                for (let i = 0; i < hierarchy.length; i++) {
                    const item = hierarchy[i];

                    item.$.filtered = true;

                    if (item.$.filtered) {
                        //filteredCount++;
                    }

                    if (item['children']) {
                        filterBy(item['children'], item, parentItem);
                    }
                }

                if (parentItem) {
                    parentItem.$.filtered = true;

                    if (root && !root.$.filtered) {
                        root.$.filtered = true;
                    }
                }
            }

            filterBy(that.boundHierarchy, null, null);
        }

        if (that.onFilter) {
            that.onFilter()
        }
    }

    clearSort() {
        const that = this;

        that._sort(that.boundSource, [], [], []);
    }

    _sort(dataSource, sortColumns, directions, dataTypes, customSortingCallback, customCompareFunctions) {
        const that = this;

        let isObservableArray = false;

        if (dataSource.length === 0) {
            return;
        }

        if (dataSource && dataSource.constructor && dataSource instanceof Smart.ObservableArray) {
            isObservableArray = true;
        }

        if (!dataSource || !(Array.isArray(dataSource)) || dataSource.length === 0 ||
            !sortColumns || Array.isArray(sortColumns) && sortColumns.length === 0) {
            if (!isObservableArray && !that.boundHierarchy) {
                throw new Error('sort: Missing or Invalid arguments!');
            }
        }

        if (typeof sortColumns === 'string') {
            sortColumns = [sortColumns];
        }

        const directionCoefficients = [],
            compareFunctions = [];

        if (directions === undefined) {
            directions = [];
        }

        const getCompareFunction = function (a, knownDataType) {
            // gets data type of column (not necessary if the Grid provides this information)
            const dataType = knownDataType || typeof a;
            let compareFunction;

            switch (dataType) {
                case 'string':
                    compareFunction = new Intl.Collator().compare;
                    break;
                case 'number':
                case 'int':
                case 'bigInt':
                case 'float':
                case 'double':
                    compareFunction = function (a, b) {
                        return a - b;
                    };
                    break;
                case 'boolean':
                case 'bool':
                    compareFunction = function (a, b) {
                        if (a === b) {
                            return 0;
                        }
                        else if (a === false) {
                            return -1;
                        }
                        else {
                            return 1;
                        }
                    };
                    break;
                case 'date':
                case 'time':
                case 'dateTime':
                    if (a instanceof Date) {
                        compareFunction = function (a, b) {
                            if (!a && !b) {
                                return 0;
                            }

                            if (!a) {
                                return -1;
                            }

                            if (!b) {
                                return 1;
                            }

                            return a.getTime() - b.getTime();
                        };
                    }
                    else if (a instanceof Smart.Utilities.DateTime ||
                        a instanceof Smart.Utilities.BigNumber) {
                        compareFunction = function (a, b) {
                            return a.compare(b);
                        };
                    }
                    break;
                case 'object':
                    if (a instanceof Date) {
                        compareFunction = function (a, b) {
                            if (!a && !b) {
                                return 0;
                            }

                            if (!a) {
                                return -1;
                            }

                            if (!b) {
                                return 1;
                            }

                            return a.getTime() - b.getTime();
                        };
                    }
                    else if (a instanceof Smart.Utilities.DateTime ||
                        a instanceof Smart.Utilities.BigNumber) {
                        compareFunction = function (a, b) {
                            return a.compare(b);
                        };
                    }
                    else if (a instanceof Smart.Utilities.Complex || (window.NIComplex && a instanceof window.NIComplex)) {
                        const complexNumericProcessor = new Smart.Utilities.ComplexNumericProcessor();

                        compareFunction = function (a, b) {
                            return complexNumericProcessor.compareComplexNumbers(a, b);
                        }
                    }

                    break;
            }

            return compareFunction;
        }

        for (let i = 0; i < sortColumns.length; i++) {
            if (directions[i] === undefined || directions[i] === 'asc' || directions[i] === 'ascending') {
                directionCoefficients[i] = 1;
            }
            else {
                directionCoefficients[i] = -1;
            }

            let value = dataSource[0][sortColumns[i]];

            if (value === undefined && dataSource[0] && dataSource[0].children && dataSource[0].data) {
                value = dataSource[0].data[sortColumns[i]];
            }

            if (customCompareFunctions && customCompareFunctions) {
                const dataField = sortColumns[i];
                if (customCompareFunctions[dataField]) {
                    compareFunctions[i] = customCompareFunctions[dataField];
                    continue;
                }
            }

            compareFunctions[i] = getCompareFunction(value, dataTypes[i]);
        }

        if (sortColumns.length > 0 && customSortingCallback) {
            customSortingCallback(dataSource, sortColumns, directions, compareFunctions);
        }
        else {
            dataSource.sort(function (a, b) {
                for (let i = 0; i < sortColumns.length; i++) {
                    const result = compareFunctions[i](a[sortColumns[i]], b[sortColumns[i]]);

                    if (result === 0) {
                        if (sortColumns[i + 1]) {
                            continue;
                        }
                        else if (a.$.index !== undefined) {
                            // makes sorting stable
                            return (a.$.index - b.$.index) * directionCoefficients[i];
                        }

                        return 0;
                    }

                    return result * directionCoefficients[i];
                }

                if (sortColumns.length === 0) {
                    if (a.$.index < b.$.index) {
                        return -1;
                    }

                    if (a.$.index > b.$.index) {
                        return 1;
                    }

                    return 0;

                }
            });
        }

        for (let i = 0; i < dataSource.length; i++) {
            that[i] = dataSource[i];
        }
    }

    static Filter(dataSource, filterColumns, filterGroups, customFilteringCallback, operator = 'and') {
        let defaultResult, operatorSpecificEval;

        if (operator === 'and') {
            defaultResult = true;
            operatorSpecificEval = function (result, dataItem, filterColumn, filterGroup) {
                if (customFilteringCallback) {
                    return result && customFilteringCallback(dataItem, filterColumn, filterGroup);
                }

                return result && filterGroup.evaluate(dataItem[filterColumn]);
            };
        }
        else {
            defaultResult = false;
            operatorSpecificEval = function (result, dataItem, filterColumn, filterGroup) {
                if (customFilteringCallback) {
                    return result || customFilteringCallback(dataItem, filterColumn, filterGroup);
                }

                return result || filterGroup.evaluate(dataItem[filterColumn]);
            };
        }

        const filteredData = dataSource.filter((dataItem) => {
            let result = defaultResult;

            for (let i = 0; i < filterGroups.length; i++) {
                const filterGroup = filterGroups[i];
                const filterColumn = filterColumns[i];

                result = operatorSpecificEval(result, dataItem, filterColumn, filterGroup);
            }

            return result;
        });

        return filteredData;
    }

    filter(filterColumns, filterGroups, customFilteringCallback) {
        Smart.DataAdapter.Filter(this.boundSource, filterColumns, filterGroups, customFilteringCallback);
    }

    sort(sortColumns, directions, customSortingCallback) {
        Smart.DataAdapter.Sort(this.boundSource, sortColumns, directions, customSortingCallback);
    }

    static Sort(dataSource, sortColumns, directions, customSortingCallback) {
        const getCompareFunction = function (a) {
            // gets data type of column (not necessary if the Grid provides this information)
            const dataType = typeof a;
            let compareFunction;

            switch (dataType) {
                case 'string':
                    if (a.indexOf('-') >= 0 || a.indexOf('/') >= 0) {
                        const testDate = new Date(a);
                        if (testDate && testDate.toString() !== 'Invalid Date') {
                            const compareFunction = function (a, b) {
                                const aDate = new Date(a);
                                const bDate = new Date(b);

                                if (!aDate && !bDate) {
                                    return 0;
                                }

                                if (!aDate) {
                                    return -1;
                                }

                                if (!bDate) {
                                    return 1;
                                }

                                return aDate.getTime() - bDate.getTime();
                            };

                            return compareFunction;
                        }
                    }

                    compareFunction = new Intl.Collator().compare;
                    break;
                case 'number':
                    compareFunction = function (a, b) {
                        return a - b;
                    };
                    break;
                case 'boolean':
                    compareFunction = function (a, b) {
                        if (a === b) {
                            return 0;
                        }
                        else if (a === false) {
                            return -1;
                        }
                        else {
                            return 1;
                        }
                    };
                    break;
                case 'object':
                    if (a instanceof Date) {
                        compareFunction = function (a, b) {
                            if (!a && !b) {
                                return 0;
                            }

                            if (!a) {
                                return -1;
                            }

                            if (!b) {
                                return 1;
                            }

                            return a.getTime() - b.getTime();
                        };
                    }
                    else if (a instanceof Smart.Utilities.DateTime ||
                        a instanceof BigNumberNG) {
                        compareFunction = function (a, b) {
                            return a.compare(b);
                        };
                    }
                    else if (a instanceof Smart.Utilities.Complex || (window.NIComplex && a instanceof window.NIComplex)) {
                        const complexNumericProcessor = new Smart.Utilities.ComplexNumericProcessor();

                        compareFunction = function (a, b) {
                            return complexNumericProcessor.compareComplexNumbers(a, b);
                        }
                    }

                    break;
            }

            return compareFunction;
        }

        if (!dataSource || !(Array.isArray(dataSource)) || dataSource.length === 0 ||
            !sortColumns || Array.isArray(sortColumns) && sortColumns.length === 0) {
            return;
        }

        if (typeof sortColumns === 'string') {
            sortColumns = [sortColumns];
        }

        const directionCoefficients = [],
            compareFunctions = [];

        if (directions === undefined) {
            directions = [];
        }

        for (let i = 0; i < sortColumns.length; i++) {
            if (directions[i] === undefined || directions[i] === 'asc' || directions[i] === 'ascending') {
                directionCoefficients[i] = 1;
            }
            else {
                directionCoefficients[i] = -1;
            }

            compareFunctions[i] = getCompareFunction(dataSource[0][sortColumns[i]]);
        }

        if (customSortingCallback) {
            customSortingCallback(dataSource, sortColumns, directions, compareFunctions);
            return;
        }

        const sortedData = dataSource.slice(0);

        sortedData.sort(function (a, b) {
            for (let i = 0; i < sortColumns.length; i++) {
                const result = compareFunctions[i](a[sortColumns[i]], b[sortColumns[i]]);

                if (result === 0) {
                    if (sortColumns[i + 1]) {
                        continue;
                    }
                    else if (a._index !== undefined) {
                        // makes sorting stable
                        return (a._index - b._index) * directionCoefficients[i];
                    }

                    return 0;
                }

                return result * directionCoefficients[i];
            }
        });

        return sortedData;
    }
}

Smart.DataAdapter = window.smartDataAdapter = DataAdapter;

class Ajax {
    constructor(config, callback) {
        const that = this;

        that.config = config;
        that.callback = callback;

        if (config.autoFetch === false) {
            return;
        }

        that.call(config);
    }

    call(config) {
        const that = this;

        if (!config) {
            config = that.config;
        }

        let method = 'GET',
            url = config.url,
            body = null,
            async = true;

        if (config.type) {
            method = config.type;
        }

        if (config.data) {
            if (method === 'GET') {
                url += '?';

                for (let prop in config.data) {
                    if (Object.prototype.hasOwnProperty.call(config.data, prop)) {
                        url += encodeURI(prop + '=' + config.data[prop] + '&');
                    }
                }

                if (url.charAt(url.length - 1) === '&') {
                    url = url.slice(0, url.length - 1);
                }
            }
            else if (method === 'POST') {
                body = JSON.stringify(config.data);
            }
        }

        if (config && config.async === false && config.dataSourceType !== 'xlsx') {
            async = false;
        }

        if (window.fetch !== undefined && async) {
            that.ajaxFetch(config, method, url, body);
        }
        else {
            that.ajaxXMLHttpRequest(config, method, url, body, async);
        }
    }

    ajaxFetch(config, method, url, body) {
        // prepare fetch config
        const that = this;
        const fetchInit = { method: method };
        let parseMethod;

        switch (config.dataSourceType) {
            case 'json':
                parseMethod = 'json';
                break;
            case 'xlsx':
                parseMethod = 'arrayBuffer';
                break;
            default:
                parseMethod = 'text';
        }

        if (config) {
            if (config.contentType) {
                fetchInit.headers = new Headers({
                    'Content-Type': config.contentType
                });
            }
        }

        if (body !== null) {
            fetchInit.body = body;
        }

        let status, fetchTimeout, timeouted;

        if (config.timeout) {
            fetchTimeout = setTimeout(function () {
                timeouted = true;
            }, config.timeout);
        }

        if (config.beforeSend) {
            const beforeSendResult = config.beforeSend(fetchInit, config);

            if (beforeSendResult === false) {
                return;
            }
        }

        // fetch resource
        fetch(url, fetchInit)
            .then(function (response) {
                if (timeouted) {
                    status = 408;
                    throw new Error('timeout');
                }

                if (fetchTimeout) {
                    clearTimeout(fetchTimeout);
                }

                status = response.status;

                if (!response.ok) {
                    throw new Error(response.statusText);
                }

                return response[parseMethod]();
            })
            .then(function (data) {
                if (parseMethod === 'arrayBuffer') {
                    return JSZip.loadAsync(data).then(function (zipData) {
                        // "data" represents the whole XLSX/ZIP file
                        return zipData.files['xl/worksheets/sheet1.xml'].async('text').then(function (sheet1) {
                            return zipData.files['xl/sharedStrings.xml'].async('text').then(function (sharedStrings) {
                                const parsedData = that.parseXLSXData(sheet1, sharedStrings);

                                that.ajaxComplete(config, parsedData, status);
                            });
                        });
                    });
                }
                else {
                    that.ajaxComplete(config, data, status);
                }
            })
            .catch(function (error) {
                if (error.message === 'JSZip is not defined') {
                    error.message = 'JSZip is not defined. Please include a reference to JSZip to be able to load data from XLSX files.';
                }

                if (config && config.loadError) {
                    config.loadError(status, error);
                }

                if (that.callback) {
                    that.callback(error, status);
                }
            });
    }

    ajaxXMLHttpRequest(config, method, url, body, async) {
        const request = new XMLHttpRequest();
        const that = this;

        request.open(method, url, async);

        request.ontimeout = function () {
            if (config && config.loadError) {
                config.loadError(408, 'timeout');
            }
        };

        request.onload = function () {
            if (request.readyState === 4) {
                const status = request.status;
                let data = request.response;

                if (status >= 200 && status <= 299) {
                    if (config.dataSourceType === 'json') {
                        data = JSON.parse(data);
                    }

                    that.ajaxComplete(config, data, status);
                }
                else if (config && config.loadError) {
                    config.loadError(status, data);
                }
            }
        };

        request.onerror = function () {
            if (config && config.loadError) {
                config.loadError(request.status, request.response);
            }
        };

        if (config && config.contentType) {
            request.setRequestHeader('Content-Type', config.contentType);
        }

        if (async && config.timeout) {
            request.timeout = config.timeout;
        }

        if (config.beforeSend) {
            const beforeSendResult = config.beforeSend(request, config);

            if (beforeSendResult === false) {
                return;
            }
        }

        request.send(body);
    }

    ajaxComplete(config, data, status) {
        if (!config) {
            return;
        }

        if (config.beforeLoadComplete) {
            const processedData = config.beforeLoadComplete(data);

            if (processedData) {
                data = processedData;
            }
        }

        if (config.loadComplete) {
            config.loadComplete(data, status);
        }

        if (this.callback) {
            this.callback(data, status);
        }
    }

    parseXLSXData(sheet1, sharedStrings) {
        const parser = new DOMParser(),
            sharedStringsDocument = parser.parseFromString(sharedStrings, 'text/xml'),
            sharedStringsContainers = Array.from(sharedStringsDocument.getElementsByTagName('si')),
            sharedStringsCollection = [],
            sheet1Document = parser.parseFromString(sheet1, 'text/xml'),
            rows = Array.from(sheet1Document.getElementsByTagName('row')),
            parsedData = [];

        sharedStringsContainers.forEach(function (si) {
            let texts = si.getElementsByTagName('t');

            if (texts.length === 1) {
                sharedStringsCollection.push(texts[0].innerHTML);
            }
            else {
                let text = '';

                texts = Array.from(texts);
                texts.forEach(function (t) {
                    text += t.innerHTML;
                });
                sharedStringsCollection.push(text);
            }
        });

        rows.forEach(function (row) {
            const rowObject = {},
                cells = Array.from(row.getElementsByTagName('c'));

            cells.forEach(function (cell/*, index*/) {
                const column = cell.getAttribute('r').match(/\D+/)[0],
                    type = cell.getAttribute('t'),
                    xmlValue = cell.getElementsByTagName('v')[0].innerHTML;
                let value;

                switch (type) {
                    case 's':
                        // string
                        value = sharedStringsCollection[parseFloat(xmlValue)];
                        break;
                    case 'b':
                        // boolean
                        value = parseFloat(xmlValue) === 1;
                        break;
                    default:
                        // number or date
                        value = parseFloat(xmlValue);
                }

                rowObject[column] = value;
            });

            parsedData.push(rowObject);
        });

        return parsedData;
    }
}

Smart.Ajax = window.Ajax = Ajax;
