
/* Smart UI v1.0.2 (2021-08-17) 
Copyright (c) 2011-2021 jQWidgets. 
License: https://htmlelements.com/license/ */ //

/**
* Gantt custom element.
*/
Smart('smart-gantt', class Gantt extends Smart.Table {
    /**
    * Gantt properties
    */
    static get properties() {
        return {
            'animation': {
                value: 'none',
                type: 'string'
            },
            'messages': {
                value: {
                    'en': {
                        'add': 'Add condition',
                        'all': 'All columns',
                        'apply': 'Apply',
                        'between': 'Between',
                        'cancel': 'Cancel',
                        'clearFilter': 'Clear filter',
                        'close': 'Close',
                        'column': 'Column:',
                        'condition': 'Condition:',
                        'conditionalFormatting': 'Conditional Formatting',
                        'CONTAINS': 'contains',
                        'CONTAINS_CASE_SENSITIVE': 'contains (case sensitive)',
                        'DOES_NOT_CONTAIN': 'does not contain',
                        'DOES_NOT_CONTAIN_CASE_SENSITIVE': 'does not contain (case sensitive)',
                        'EMPTY': 'empty',
                        'ENDS_WITH': 'ends with',
                        'ENDS_WITH_CASE_SENSITIVE': 'ends with (case sensitive)',
                        'EQUAL': 'equal',
                        'equal': 'Equal To',
                        'EQUAL_CASE_SENSITIVE': 'equal (case sensitive)',
                        'filterCondition': 'Filter condition',
                        'filterPlaceholder': 'Filter',
                        'firstButton': 'First',
                        'fontFamily': 'Font family:',
                        'fontSize': 'Font size:',
                        'format': 'Format:',
                        'formatColumn': 'Format Column',
                        'GREATER_THAN': 'greater than',
                        'GREATER_THAN_OR_EQUAL': 'greater than or equal',
                        'greaterThan': 'Greater Than',
                        'highlight': 'Highlight',
                        'invalidValue': 'Invalid value',
                        'itemsPerPage': 'Items per page:',
                        'lastButton': 'Last',
                        'LESS_THAN': 'less than',
                        'LESS_THAN_OR_EQUAL': 'less than or equal',
                        'lessThan': 'Less Than',
                        'nextButton': 'Next',
                        'NOT_EMPTY': 'not empty',
                        'NOT_EQUAL': 'not equal',
                        'NOT_NULL': 'not null',
                        'notEqual': 'Not Equal To',
                        'NULL': 'null',
                        'ok': 'OK',
                        'previousButton': 'Previous',
                        'remove': 'Remove condition',
                        'secondValue': 'Second value:',
                        'STARTS_WITH': 'starts with',
                        'STARTS_WITH_CASE_SENSITIVE': 'starts with (case sensitive)',
                        'summaryPrefix': 'of',
                        'text': 'Text',
                        'value': 'Value:',
                        'with': 'with',
                        'grouping': 'Group field name is not found in the dataFields collection!',
                        'invalidTicketTemplate': 'The "{{property}}" property of the tickets is not an HTMLTemplate element.',
                        'unassign': 'Unassign',
                        'editRoute': 'Edit Route',
                        'eod': 'EOD',
                        'preview': 'Preview',
                        'submit': 'Submit',
                        'undo': 'Undo',
                        'redo': 'Redo',
                        'seeOriginal': 'See Original',
                    }
                },
                type: 'object',
                extend: true
            },
            'rowDetailTemplate': {
                value: '<smart-table class="smart-gantt-row-detail-table" virtualization wait></smart-table>',
                type: 'string?'
            },
            'virtualization': {
                readOnly: true,
                value: true,
                type: 'boolean'
            }
        }
    }

    /**
     * Gantt event listeners.
     */
    static get listeners() {
        return {
            'tableContainer.click': '_tableContainerClickHandler',
        };
    }

    /**
     * Gantt template
     * @returns 
     */
    template() {
        return `<div id="container" class="smart-container" role="presentation">
                        <div id="header" class="smart-table-header" role="toolbar">
                            <smart-input id="filterInput" class="underlined" animation="[[animation]]" locale="[[locale]]" right-to-left="[[rightToLeft]]" theme="[[theme]]"></smart-input>
                            <div id="filterTemplateContainer" class="smart-table-filter-template-container smart-hidden"></div>
                            <smart-button id="conditionalFormattingButton" class="smart-table-toolbar-button conditional-formatting" animation="[[animation]]" right-to-left="[[rightToLeft]]" theme="[[theme]]" aria-label="Conditional Formatting"></smart-button>
                        </div>
                        <smart-scroll-viewer style="opacity: 0.99" id="virtualizationContainer" class="smart-table-virtualization-container" right-to-left="[[rightToLeft]]" theme="[[theme]]">
                            <table id="tableContainer" inner-h-t-m-l=\'[[innerHTML]]\' class="smart-table-container">
                                <content></content>
                            </table>
                        </smart-scroll-viewer>
                        <smart-pager id="pager" animation="[[animation]]" locale="[[locale]]" page-index="[[pageIndex]]" page-size="[[pageSize]]" pages-count="null" show-first-last-navigation-buttons show-page-size-selector show-prev-next-navigation-buttons show-summary right-to-left="[[rightToLeft]]" theme="[[theme]]"></smart-pager>
                        <div id="loadingIndicatorContainer" class="smart-loader-container smart-hidden" aria-label="Loading">
                            <span id="loadingIndicator" class="smart-loader" role="presentation"></span>
                        </div>
                    </div>`
    }

    /**
     * Gantt render method
     */
    _render() {
        const that = this,
            computedStyle = getComputedStyle(that);

        that.setAttribute('role', 'presentation');
        that.$.tableContainer.id = that.id + 'TableContainer';

        that._cachedWidth = that.offsetWidth;
        that._cachedHeight = that.offsetHeight;
        that._defaults = {
            fontFamily: computedStyle.fontFamily,
            fontSize: computedStyle.fontSize,
            text: that._toHex(computedStyle.color),
            highlight: that._toHex(computedStyle.backgroundColor)
        };
        that._isMobile = Smart.Utilities.Core.isMobile;
        that._autoScrollCoefficient = Smart.Utilities.Core.Browser.Firefox ? 8 : Smart.Utilities.Core.Browser.Edge ? 16 : 4;
        that._dblclickObject = { numberOfClicks: 0 };
        that._expandedIds = [];
        that._selectedIds = [];
        that._disabledSelection = [];
        that._selectableGroupRecords = [];
        that._state = { expanded: [], selected: [], sorted: [] };
        that._rowHeight = parseFloat(computedStyle.getPropertyValue('--smart-gantt-row-height'));
        that._rowDetailHeight = parseFloat(computedStyle.getPropertyValue('--smart-gantt-row-detail-height'));
        that._rowEditHeight = parseFloat(computedStyle.getPropertyValue('--smart-gantt-row-edit-height'));
        that._animation = that.animation;

        that._setFocusable();

        if (that.selection) {
            that.$.tableContainer.setAttribute('aria-multiselectable', true);
        }

        if (that.virtualization) {
            const virtualizationContainer = that.$.virtualizationContainer;
            const table = that;

            if (virtualizationContainer) {
                virtualizationContainer._mouseWheelHandler = function (event) {
                    const that = virtualizationContainer;

                    if (that.disabled || (!that.computedHorizontalScrollBarVisibility && !that.computedVerticalScrollBarVisibility)) {
                        return;
                    }

                    if (event.shiftKey && that.computedHorizontalScrollBarVisibility) {
                        const scrollLeft = that.scrollLeft;

                        if (scrollLeft === 0 && event.deltaX < 0 ||
                            scrollLeft === that.scrollHeight && event.deltaX > 0) {
                            return;
                        }
                        event.stopPropagation();
                        event.preventDefault();

                        if (that.scrollWidth > 0) {
                            that.scrollTo(undefined, that.scrollLeft + that._getScrollCoefficient(event, that.offsetWidth));
                        }

                        return;
                    }

                    if (that.computedVerticalScrollBarVisibility) {
                        const scrollTop = that.scrollTop;

                        if (scrollTop === 0 && event.deltaY < 0 ||
                            scrollTop === that.scrollHeight && event.deltaY > 0) {
                            return;
                        }

                        event.stopPropagation();
                        event.preventDefault();

                        if (that.scrollHeight > 0) {
                            let verticalOffset = 3 * table._rowHeight;

                            if (event.deltaY <= 0) {
                                verticalOffset = -3 * table._rowHeight;
                            }

                            if (Math.abs(event.deltaY) >= 100) {
                                that.scrollTop += verticalOffset;
                            }
                            else {
                                that.scrollTop += event.deltaY;
                            }
                        }
                    }
                }

                virtualizationContainer._verticalScrollbarHandler = function (event) {
                    if (that.deferredScrollDelay === 0) {
                        that._onVerticalChange(event);
                    }
                    else {
                        if (that._verticalScrollTimer) {
                            clearTimeout(that._verticalScrollTimer);
                            delete that._verticalScrollTimer;
                        }
                        that._verticalScrollTimer = setTimeout(() => {
                            that._onVerticalChange(event);
                            delete that._verticalScrollTimer;
                        }, that.deferredScrollDelay);
                    }
                }
            }

            virtualizationContainer.hasStyleObserver = false;
            virtualizationContainer.$.verticalScrollBar.hasStyleObserver = false;
            virtualizationContainer.$.horizontalScrollBar.hasStyleObserver = false;

            that.freezeHeader = true;
            that.freezeFooter = true;
        }

        if (that.isRendered) {
            if (that.innerHTML && that.innerHTML.trim().length > 1) {
                that._initDataSourceFromHTML();
            }
            else {
                that._fullRefresh();
            }

            return;
        }

        that._createElement();

        // update row height after a small amount of time. This is required, if images are present within Table cells and loaded with delay.
        setTimeout(() => {
            const renderedRows = that.querySelectorAll('tr[row-id]');

            if (renderedRows.length > 0) {
                let isLayoutDirty = false;

                for (let i = 0; i < renderedRows.length; i++) {
                    const firstRow = renderedRows[i];

                    if (that._rowHeight < firstRow.offsetHeight) {
                        that._rowHeight = firstRow.offsetHeight;
                        isLayoutDirty = true;
                    }
                }

                if (isLayoutDirty) {
                    that.style.setProperty('--smart-table-row-height', that._rowHeight + 'px');
                    that._refreshDataRows();
                }
            }
        }, 500);
    }

    /**
     * Called when the element is attached to the DOM.
     */
    attached() {
        const that = this;

        super.attached();

    }


    /**
     * Called when the element is detached from the DOM.
     */
    detached() {
        const that = this;

        super.detached();
    }

    /**
     * Expands the details of a row
     */
    beginRowEdit(rowId) {
        const that = this;

        if (!that.rowDetailTemplate) {
            return
        }

        const rowObject = that.rowById ? that.rowById[rowId] : that.rows.dataItemById[rowId],
            data = rowObject.data || rowObject;

        if (!rowObject || !data) {
            return;
        }

        const rowsInDOM = that._rowsInDOM;
        let row;

        for (let i = 0, max = rowsInDOM.length; i < max; i += 1) {
            const r = rowsInDOM[i];

            if (r.data && r.data.$.id === rowId && !r.classList.contains('smart-table-row-detail') && !r.hasAttribute('unused')) {
                row = r;
                break;
            }
        }

        if (!row) {
            return
        }

        const detailRow = row.nextElementSibling;

        if (!detailRow && !that.virtualization || detailRow && detailRow.classList.contains('smart-table-row-detail') && !detailRow.classList.contains('collapsed')) {
            return
        }

        let rowEditing = that._rowEditing;

        if (!rowEditing) {
            rowEditing = that._rowEditing = [];
        }

        if (rowEditing.indexOf(rowId) > -1) {
            return
        }

        rowEditing.push(rowId);

        row.setAttribute('edit', '');

        that._toggleRowDetail(row, data.$.id);
    }

    endRowEdit(rowId) {
        const that = this;

        if (!that.rowDetailTemplate) {
            return
        }

        const rowObject = that.rowById ? that.rowById[rowId] : that.rows.dataItemById[rowId],
            data = rowObject.data || rowObject;

        if (!rowObject || !data) {
            return;
        }

        const rowsInDOM = that._rowsInDOM;
        let row;

        for (let i = 0, max = rowsInDOM.length; i < max; i += 1) {
            const r = rowsInDOM[i];

            if (r.data && r.data.$.id === rowId && !r.classList.contains('smart-table-row-detail') && !r.hasAttribute('unused')) {
                row = r;
                break;
            }
        }

        if (!row) {
            return
        }

        const detailRow = row.nextElementSibling;

        if (!detailRow || !detailRow.classList.contains('smart-table-row-detail') || detailRow.classList.contains('collapsed')) {
            return
        }

        const rowEditing = that._rowEditing;

        if (!rowEditing || rowEditing.indexOf(rowId) < 0) {
            return
        }

        rowEditing.splice(rowEditing.indexOf(rowId), 1);

        row.removeAttribute('edit');

        that._hideRowDetail(row, data.$.id);
    }

    /**
     * Refreshes the UI Component.
     */
    refresh() {
        const that = this;

        if (that.__updating) {
            return;
        }

        const columns = that._columns.filter(col => col.visible),
            tableContainer = that.$.tableContainer,
            filterRow = tableContainer.querySelector('.smart-table-filter-row'),
            footer = tableContainer.querySelector('tfoot'),
            tbody = document.createElement('tbody'),
            fractionOfMax = that._getFractionOfMax(),
            scrollContainer = that.virtualization ? that.$.virtualizationContainer : that.$.tableContainer.parentElement,
            currentScroll = { top: scrollContainer.scrollTop, left: scrollContainer.scrollLeft };

        that.columns.canNotify = false;

        while (tableContainer.childNodes.length > 0) {
            tableContainer.childNodes[0].remove();
        }

        that._createTableHeader();

        if (!that.isRendered && that.onInit && !that._loadingState) {
            that.onInit();
        }

        that._createCustomHeaderRow();
        that._createFooterRow(footer);
        that._createDataRows(tbody);
        that._createFilterRow(filterRow);

        //Refresh the tickets
        that._recycleRowTickets();

        const headerCellElements = tableContainer.querySelectorAll('th[data-field]');

        for (let i = 0; i < columns.length; i++) {
            const column = columns[i];

            column.headerCellElement = headerCellElements[i];

            column.headerCellElement.onclick = function () {
                const dataField = this.dataField;

                that.$.fireEvent('columnClick', { dataField: dataField });

                if (that.sortMode === 'none' || column.allowSort === false || column.dataField !== 'timeline' && that._preventClickSort) {
                    return;
                }

                that._addSortIconContainer(this);

                that._sortOnClick = true;

                if (this.headerCellElement.sortIconContainerElement.classList.contains('asc')) {
                    that.sortBy(dataField, 'desc');
                }
                else if (this.headerCellElement.sortIconContainerElement.classList.contains('desc')) {
                    that.sortBy(dataField, null);
                }
                else {
                    that.sortBy(dataField, 'asc');
                }

                that._sortOnClick = false;
            }.bind(column);
        }

        if (that._sortColumns && that._sortColumns.length > 0) {
            for (let i = 0; i < that._sortColumns.length; i++) {
                const sortColumn = that._sortColumns[i];
                const column = that.columnByDataField[sortColumn.dataField];

                that._addSortIconContainer(column, sortColumn.direction, i + 1);
            }
        }

        that._handleAutoSizeMode();
        that._handleFrozenColumnPositions();
        that._refreshHorizontalScrollbar();

        if (that.isRendered) {
            if (that.selection) {
                that._updateSelectAllState();
            }

            that._setFractionOfMax(fractionOfMax);
        }

        //Reset the scrolling again
        [scrollContainer.scrollTop, scrollContainer.scrollLeft] = [currentScroll.top, currentScroll.left];

        that.columns.canNotify = true;
    }

    sortBy(columnDataField, sortOrder) {
        const that = this,
            column = that.columnByDataField[columnDataField];

        if (column) {
            if (column.allowSort === false) {
                return;
            }

            columnDataField = column.dataField;
        }

        //NOTE: Cant sort by the 'timeline' column
        if (columnDataField === 'timeline') {
            return
        }

        that._sortBy({
            column: column,
            columnDataField: columnDataField,
            sortOrder: sortOrder,
            dataFields: that.dataSource.dataFields,
            columnByDataField: 'columnByDataField'
        });

        if (columnDataField === null) {
            const eventType = that._sortOnClick === true ? 'interaction' : 'programmatic';

            that.$.fireEvent('sort', { columns: [], sortDataFields: [], sortOrders: [], sortDataTypes: [], type: eventType });
        }
    }

    /**
     * Handles the context menu open/close
     * @param {Event} event 
     */
    _handleRowMenu(target) {
        const that = this,
            row = target.closest('tr');

        if (row && that._rowsInDOM.indexOf(row) > -1) {
            const rowId = row.data ? row.data.$.id : undefined;

            if (rowId === undefined || rowId === null) {
                return
            }

            let menu = that._menu;

            if (!menu) {
                menu = that._menu = document.createElement('smart-menu');
                menu.classList.add('smart-gantt-menu');
                menu.mode = 'dropDown';
            }

            menu.addEventListener('itemClick', that._handleMenuItemClick.bind(that, rowId), { once: true });
            menu.rightToLeft = that.rightToLeft;
            menu.animation = that.animation;
            menu.theme = that.theme;

            if (!menu.dataSource) {
                menu.dataSource = [{
                    label: that.localize('editRoute'),
                    value: 'editRoute'
                },
                {
                    label: that.localize('eod'),
                    value: 'eod'
                }];
            }

            if (!menu.parentElement) {
                that.getShadowRootOrBody().appendChild(menu);
            }

            const targetRect = target.getBoundingClientRect();

            menu.open(targetRect.left, targetRect.top);
        }
    }

    /**
      * Creates row detail on demand.
      */
    _createRowDetailOnDemand(rowElement, rowObject, data) {
        const that = this,
            colspan = Number(that.selection) + Number(!!that.rowDetailTemplate) + that._columns.filter(col => col.visible).length;

        rowElement.data = Object.assign(data, { detail: true });
        rowElement.classList.add('smart-table-row-detail');
        rowElement.removeAttribute('aria-rowindex');
        rowElement.removeAttribute('row-id');

        if (!rowElement.hasAttribute('unused')) {
            rowElement.innerHTML = `<td colspan="${colspan}"><div class="smart-table-detail-container">${that._applyRowDetailTemplate(rowObject)}</div></td>`;
        }

        that._setRowDetail(rowObject, rowElement);
    }

    /**
     * Creates Table data rows when virtualization is enabled.
     */
    _createVirtualDataRows(tbody) {
        const that = this,
            rowDetailTemplate = that.rowDetailTemplate,
            selection = that.selection,
            hierarchical = that.hasAttribute('hierarchy'),
            fragment = document.createDocumentFragment(),
            virtualizationContainer = that.$.virtualizationContainer,
            tableContainer = that.$.tableContainer,
            rowHeight = that._rowHeight,
            rowEditHeight = that._rowEditHeight;

        if (that.isRendered) {
            virtualizationContainer.scrollWidth = 0;
            virtualizationContainer.scrollHeight = 0;
            virtualizationContainer.$.scrollViewerContainer.classList.remove('hscroll', 'vscroll');
        }

        const viewHeight = virtualizationContainer.$.scrollViewerContainer.offsetHeight - tableContainer.firstElementChild.offsetHeight -
            (that.filtering && that.filterRow ? rowHeight : 0) -
            (that.footerRow ? tableContainer.lastElementChild.offsetHeight : 0),
            tableAutoHeight = viewHeight < 5, // height of Table is close to height of header => height of Table is 'auto'
            rowsVirtual = that._getVirtualRecords(hierarchical, tableAutoHeight),
            numberOfVirtualRows = rowsVirtual.length,
            rowEditing = that._rowEditing,
            editingRowsCount = rowEditing ? rowEditing.length : 0;
        let maxHeight = hierarchical ? viewHeight * 2 : viewHeight + 3 * rowHeight,
            scrollHeight = rowHeight * (numberOfVirtualRows - editingRowsCount) + editingRowsCount * rowEditHeight,
            accumulatedHeight = 0;

        that._rowsVirtual = rowsVirtual;
        that._rowsInDOM = [];

        for (let i = 0; i < numberOfVirtualRows; i++) {
            const rowObject = rowsVirtual[i],
                data = rowObject.data;

            if (rowObject.detail) {
                continue;
            }

            if (data.groupDataField) {
                that._rowsInDOM.push(that._createGroupHeader(data, fragment));
            }
            else {
                const newDataRow = that._createDataRow(
                    rowObject,
                    fragment,
                    {
                        hierarchical: hierarchical, rowDetailTemplate: rowDetailTemplate,
                        selection: selection, virtualization: true
                    }
                );

                that._rowsInDOM.push(newDataRow.row);

                if (newDataRow.detail) {
                    that._rowsInDOM.push(newDataRow.detail);
                    accumulatedHeight += that._rowDetailHeight;
                    maxHeight += that._rowDetailHeight;
                }
            }

            accumulatedHeight += rowHeight;

            if (!tableAutoHeight && accumulatedHeight > maxHeight) {
                // do not create more rows than can be shown in the view + necessary "padding"
                break;
            }
        }

        tbody.appendChild(fragment);
        that._updateState('expanded');

        if (!tbody.parentElement) {
            that.$.tableContainer.insertBefore(tbody, that.$.tableContainer.querySelector('tfoot'));
        }

        //Refresh the Tickets scroll
        that._refreshTicketScroll();

        virtualizationContainer.refresh();

        if (rowDetailTemplate) {
            const rowDetailHeightCorrection = that._expandedRowDetailIds.length * (that._rowDetailHeight - rowHeight);

            scrollHeight += rowDetailHeightCorrection;
        }

        let verticalScrollBarMax = Math.max(0, scrollHeight - viewHeight);

        if (verticalScrollBarMax > 0 &&
            virtualizationContainer.computedHorizontalScrollBarVisibility) {
            verticalScrollBarMax += virtualizationContainer.$.horizontalScrollBar.offsetHeight;
        }

        virtualizationContainer.$.scrollViewerContentContainer.style.top = 0;
        virtualizationContainer.$.verticalScrollBar.max = verticalScrollBarMax;

        that._scrollHeight = scrollHeight;
        that._start = { view: 0, data: 0 };

        if (!tableAutoHeight && that._rowsInDOM.length > 0 &&
            virtualizationContainer.$.scrollViewerContentContainer.offsetHeight <
            virtualizationContainer.$.scrollViewerContainer.offsetHeight) {
            that._rowsInDOM[that._rowsInDOM.length - 1].classList.add('last-row');
        }

        that._focusEquivalentCell();
    }

    /**
    * Creates a single Table data row.
    */
    _createDataRow(rowObject, fragment, tableSettings) {
        const that = this,
            columns = that._columns.filter(col => col.visible),
            data = rowObject.data,
            id = data.$.id,
            rowElement = document.createElement('tr'),
            result = { row: rowElement, detail: null },
            rowEditing = that._rowEditing;
        let colspan = 0;

        rowElement.setAttribute('aria-rowindex', data.$.index + 1);

        that._processAndRenderSelection(data, rowElement, tableSettings.hierarchical);

        that._processAndRenderDetails(data, rowElement, tableSettings.hierarchical);

        if (tableSettings.selection) {
            colspan++;
        }

        if (that.rowDetailTemplate) {
            colspan++;
        }

        for (let j = 0; j < columns.length; j++) {
            const td = document.createElement('td'),
                column = columns[j];

            td.setAttribute('data-field', column.dataField);
            td.setAttribute('aria-colindex', j + 1 + (+tableSettings.selection));

            let value = that._formatCellValue(rowObject, column, td, column.dataField === 'progress' ?
                { completed: data[column.dataField], total: Array.isArray(data.timeline) ? data.timeline.length : 0 } : undefined), isTreeCell;

            if (column.freeze) {
                td.classList.add('freeze-' + (column.freeze === 'far' ? 'far' : 'near'));
            }

            if (column.responsivePriority) {
                td.classList.add('priority-' + column.responsivePriority);
            }

            if (column.allowEdit === false) {
                td.classList.add('no-edit');
                td.setAttribute('aria-readonly', true);
            }

            if (j === 0 && tableSettings.hierarchical) {
                const childrenFiltered = that._areChildrenFiltered(data);

                if (!data.leaf) {
                    td.classList.add('tree-cell');
                    isTreeCell = childrenFiltered;
                }

                if (data.level) {
                    td.classList.add('outline-level-' + Math.min(data.level, 10));

                    if (data.leaf || !childrenFiltered) {
                        td.classList.add('tree-leaf');
                    }
                }
            }

            if (tableSettings.virtualization && column.templateElement) {
                const templateElement = column.templateElement + '';

                if (isTreeCell) {
                    that._setCellContent(td, value, isTreeCell, id);
                }

                const templateChild = !isTreeCell ? td : td.children[0].children[1];

                if (templateElement.indexOf('{{value}}') >= 0) {
                    templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                }
                else {
                    templateChild.innerHTML = templateElement;
                }

                if (column.templateElementSettings) {
                    column.templateElementSettings(data[column.dataField], id, templateChild.firstElementChild);
                }
            }
            else {
                that._setCellContent(td, value, isTreeCell, id);
            }

            if (that.onCellRender) {
                that.onCellRender(data, column.dataField, data[column.dataField], td);
            }

            that._applyConditionalFormattingToCell(td, column.dataField, data.$.index);

            rowElement.appendChild(td);
            colspan++;
        }

        rowElement.data = data;
        rowElement.setAttribute('row-id', id);

        if (that._disabledSelection.indexOf(id) !== -1) {
            rowElement.removeAttribute('aria-selected');
            rowElement.classList.add('disable-select');

            if (tableSettings.selection) {
                rowElement.firstElementChild.firstElementChild.setAttribute('aria-disabled', true);
            }
        }

        if (data.expanded) {
            if (that._expandedIds.indexOf(id) === -1) {
                that._expandedIds.push(id);
            }

            rowElement.setAttribute('aria-expanded', true);
            rowElement.classList.add('expanded');
        }
        else if (data.expanded === false) {
            rowElement.setAttribute('aria-expanded', false);
        }

        if (data.level && that._isCollapsed(data)) {
            rowElement.setAttribute('aria-hidden', true);
            rowElement.classList.add('collapsed', 'smart-hidden');
        }

        if (rowEditing && rowEditing.indexOf(id) > -1) {
            rowElement.setAttribute('edit', '');
        }
        else {
            rowElement.removeAttribute('edit');
        }

        fragment.appendChild(rowElement);

        //Row detail template
        if (that._expandedRowDetailIds.indexOf(id) === -1) {
            return result;
        }

        const detailTr = document.createElement('tr');

        detailTr.data = Object.assign(data, { detail: true });
        detailTr.classList.add('smart-table-row-detail');
        detailTr.innerHTML = `<td colspan="${colspan}"><div class="smart-table-detail-container">${that._applyRowDetailTemplate(rowObject)}</div></td>`;

        that._setRowDetail(rowObject, detailTr);

        fragment.appendChild(detailTr);

        result.detail = detailTr;

        return result;
    }

    /**
     * Creates the Header Timeline
     */
    _createHeaderTimeline() {
        const that = this,
            timeline = that._timeline;

        if (!timeline || !timeline.element) {
            return
        }

        const scaleCount = that._getTimelineScaleCount(),
            scaleWidth = 50; // TODO: Create a CSS var or JS property for the width of the scale cells
        let cellsObjs = [], left = 0, hour = 0;

        for (let i = 0; i < scaleCount; i += 1) {
            cellsObjs.push({
                hour: hour,
                width: scaleWidth,
                left: left
            });

            hour = (hour + 1) % 24;
            left += scaleWidth;
        }

        timeline.cellObjs = cellsObjs;
        timeline.scrollWidth = Math.max(scaleCount * scaleWidth, timeline.scrollView.offsetWidth);

        //Set size to the timeline
        // timelineElement.style.width = Math.max(scaleCount * scaleWidth, timelineElement.parentElement.offsetWidth) + 'px';
        that.$.tableContainer.style.setProperty('--smart-gantt-timeline-width', timeline.scrollWidth + 'px');

        that._recycleTimeline();
        that._refreshScrollButtons();
    }

    /**
     * Document down handler.
     */
    _documentDownHandler(event) {
        const that = this,
            menu = that._menu,
            target = that.isInShadowDOM ? event.originalEvent.composedPath()[0] : event.originalEvent.target;

        function handleDblclick() {
            if (that.columnSizeMode === 'default') {
                return;
            }

            that._preventClickSort = true;

            clearTimeout(that._dblclickObject.timeout);
            that._dblclickObject.numberOfClicks++;

            if (that._dblclickObject.numberOfClicks === 2) {
                that._dblclickObject.numberOfClicks = 0;
                that._autoResizeColumnOnDblclick();
            }
            else {
                that._dblclickObject.timeout = setTimeout(function () {
                    that._dblclickObject.numberOfClicks = 0;
                }, 300);
            }
        }

        delete that._preventClickSort;

        if (menu && menu.opened && !menu.contains(target)) {
            menu.close();

            if (that.animation !== 'none') {
                menu.ontransitionend = function () {
                    menu.ontransitionend = null;
                    menu.remove();
                }
            }
            else {
                menu.remove();
            }
        }

        if (that._editing && !target.closest('.smart-table-editor')) {
            const dropDown = target.closest('.smart-drop-down') || target.closest('.smart-input-drop-down-menu');

            if (!dropDown) {
                that.endEdit();
                return;
            }

            const ownerElement = dropDown.ownerElement;

            if (!ownerElement) {
                that.endEdit();
                return;
            }

            if (that.$.container.contains(ownerElement) && ownerElement.classList.contains('smart-table-editor')) {
                return;
            }

            that.endEdit();
            return;
        }

        const columnReorder = that.columnReorder,
            columnResize = that.columnResize;

        if (columnReorder && !that.columnTotals || columnResize) {
            const th = target.closest('th[data-field]:not(.smart-pivot-table-total-header)');

            if (th && that.$.tableContainer.contains(th)) {
                that.$.tableContainer.classList.add('smart-unselectable');

                if (columnResize && that._columns.length > 1 &&
                    !that.hasAttribute('resize-max-reached')) {
                    const rightToLeft = that.rightToLeft,
                        rect = th.getBoundingClientRect();

                    if (!rightToLeft && event.clientX <= rect.left + 5 ||
                        rightToLeft && event.clientX >= rect.right - 5) {
                        const previousItem = th.previousElementSibling;

                        if (previousItem && previousItem.hasAttribute('data-field')) {
                            if (previousItem.hasAttribute('locked')) {
                                return;
                            }

                            const previousItemRect = previousItem.getBoundingClientRect();

                            that._resizeDetails = {
                                FeedbackShown: false,
                                Item: previousItem,
                                ItemStart: previousItemRect[rightToLeft ? 'right' : 'left'] + window.scrollX,
                                Min: that._getNumericMinWidth(),
                                StartPosition: { left: event.pageX, top: event.pageY },
                                StartTime: new Date()
                            };
                            that._resizeDetails.Max = that._getCurrentResizeMax();
                            handleDblclick();
                            return;
                        }
                    }

                    if (!rightToLeft && event.clientX >= rect.right - 5 ||
                        rightToLeft && event.clientX <= rect.left + 5) {
                        if (th.hasAttribute('locked')) {
                            return;
                        }

                        that._resizeDetails = {
                            FeedbackShown: false,
                            Item: th,
                            ItemStart: rect[rightToLeft ? 'right' : 'left'] + window.scrollX,
                            Min: that._getNumericMinWidth(),
                            StartPosition: { left: event.pageX, top: event.pageY },
                            StartTime: new Date()
                        };
                        that._resizeDetails.Max = that._getCurrentResizeMax();
                        handleDblclick();
                        return;
                    }
                }

                if (columnReorder) {
                    that._dragDetails = {
                        FeedbackShown: false,
                        Item: th,
                        StartPosition: { left: event.pageX, top: event.pageY },
                        StartTime: new Date()
                    };
                }
            }
        }
    }

    /**
     * Virtualization container vertical scrolling handler.
     */
    _onVerticalChange(event, toggle) {
        const that = this;

        if (that._ignoreVerticalChange) {
            return;
        }

        const rows = that.rows,
            rowHeight = that._rowHeight,
            rowEditHeight = that._rowEditHeight,
            virtualizationContainer = that.$.virtualizationContainer;
        let value = event.detail.value;

        if (Smart.Utilities.Core.Browser.Firefox && that.virtualization) {
            virtualizationContainer.$.scrollViewerContentContainer.style.top = null;
        }

        that.setAttribute('ignore-detail-animation', '');

        if (that._rowsInDOM.length === rows.length ||
            virtualizationContainer.$.verticalScrollBar.max === 0) {
            // no virtual items
            const start = Math.floor(value / rowHeight);

            that._start = { view: start, data: start };
            virtualizationContainer.$.scrollViewerContentContainer.style.top = -value + 'px';
            requestAnimationFrame(() => that.removeAttribute('ignore-detail-animation'));
            that._fixFrozenHeaderFirefox();
            return;
        }

        const rowEditing = that._rowEditing;
        let startOfView, startOfData;

        if (that._expandedRowDetailIds.length > 0) {
            // when there are expanded row details
            let height = 0,
                difference;
            startOfView = 1;

            for (let i = 0; i < that._rowsVirtual.length; i++) {
                const previousHeight = height;

                if (that._rowsVirtual[i].detail) {
                    height += that._rowDetailHeight;
                }
                else if (rowEditing && rowEditing.indexOf(that._rowsVirtual[i].data.$.id) > -1) {
                    height += rowEditHeight;
                }
                else {
                    height += that._rowHeight;
                }

                if (startOfData === undefined && value <= height) {
                    startOfData = i;
                    difference = value - previousHeight;
                    break;
                }
            }

            //NOTE: Scrolls only when the vScroll is visible. This should only happend when the Table height is not 'auto'
            if (virtualizationContainer.computedVerticalScrollBarVisibility) {
                value = difference + that._rowHeight;
            }
            else {
                //Resets the start view item index to 0 because there was not vScroll
                startOfView = 0;
            }
        }
        else if (rowEditing && rowEditing.length > 0) {
            // when there are edit rows
            let height = 0,
                difference;
            startOfView = 1;

            for (let i = 0; i < that._rowsVirtual.length; i++) {
                const previousHeight = height,
                    rowId = that._rowsVirtual[i].data.$.id;

                if (rowEditing.indexOf(rowId) > -1) {
                    height += rowEditHeight;
                }
                else {
                    height += that._rowHeight;
                }

                if (startOfData === undefined && value <= height) {
                    startOfData = i;
                    difference = value - previousHeight;
                    break;
                }
            }

            //NOTE: Scrolls only when the vScroll is visible. This should only happend when the Table height is not 'auto'
            if (virtualizationContainer.computedVerticalScrollBarVisibility) {
                value = difference + that._rowHeight;
            }
            else {
                //Resets the start view item index to 0 because there was not vScroll
                startOfView = 0;
            }
        }
        else {
            value = value % (2 * rowHeight) + rowHeight;

            if (value < rowHeight) {
                value += rowHeight;
            }

            while (value + virtualizationContainer.$.scrollViewerContainer.offsetHeight > that.$.tableContainer.offsetHeight) {
                value -= rowHeight;
            }

            startOfView = Math.floor(value / rowHeight);
            startOfData = Math.floor(event.detail.value / rowHeight);

            if (startOfView < 0) {
                startOfData -= startOfView;
                startOfView = 0;
            }
        }

        virtualizationContainer.$.scrollViewerContentContainer.style.top = -value + 'px';

        that._updateVisibleRows(startOfView, startOfData, toggle || that._expandedRowDetailIds.length > 0);

        that._fixFrozenHeaderFirefox();

        requestAnimationFrame(() => that.removeAttribute('ignore-detail-animation'));
    }

    /**
     * Creates/Refreshes the timeline scale cells
     * @returns 
     */
    _recycleTimeline() {
        const that = this,
            timeline = that._timeline;

        if (!timeline || !timeline.element) {
            return
        }

        const timelineElement = timeline.element,
            cellObjs = timeline.cellObjs,
            { firstCellObj } = that._getVisibleCellObjRange(),
            cellsCount = that._getVisibleCellsCount(firstCellObj),
            fragment = document.createDocumentFragment(),
            timelineCells = timelineElement.children,
            firstCellObjIndex = cellObjs.indexOf(firstCellObj);

        //Remove unnecessary scale cells
        while (timelineCells.length > cellsCount) {
            timelineCells[0].remove();
        }

        //Create and update the scale cells
        for (let i = 0, max = cellsCount; i < max; i += 1) {
            const cell = timelineCells[i] || document.createElement('div'),
                cellObj = cellObjs[firstCellObjIndex + i];

            if (!cellObj) {
                break
            }

            const cellHour = cellObj.hour,
                cellStyle = cell.style;

            cellStyle.width = cellObj.width + 'px';
            cellStyle.left = cellObj.left + 'px';
            cell.classList.add('smart-timeline-cell');
            cell.setAttribute('hour', cellHour);

            cellHour === 12 ? cell.setAttribute('driver-clockout', '') : cell.removeAttribute('driver-clockout');

            if (!cell.parentElement) {
                fragment.appendChild(cell);
            }
        }

        //Add cells to the Timeline
        if (fragment.childElementCount) {
            timelineElement.appendChild(fragment);
        }
    }

    /**
     * Refreshes the Tickets
     */
    _refreshTickets(tickets, cell, rowId) {
        const that = this,
            rowEditing = that._rowEditing,
            timeline = that._timeline;

        if (!timeline || !timeline.cellObjs || !timeline.cellObjs.length) {
            if (cell) {
                cell.textContent = '';
            }
            return
        }

        //Row editing
        if (rowEditing && rowEditing.indexOf(rowId) > -1 && !cell.classList.contains('.smart-gantt-timeline-row-edit')) {
            cell.innerHTML = `<div class="smart-gantt-timeline-row-edit">
                                <div class="smart-gantt-timeline-row-header">
                                    <div>
                                        <div action="undo">${that.localize('undo')}</div>
                                        <div action="redo">${that.localize('redo')}</div>
                                        <div action="seeOriginal">${that.localize('seeOriginal')}</div>
                                    </div>
                                    <div>
                                        <div action="open"></div>
                                        <div action="close"></div>
                                    </div>
                                </div>
                                <div class="smart-gantt-timeline-row">
                                    <div class="smart-gantt-menu-button"></div><div class="smart-gantt-tickets"><div class="smart-gantt-ticket-container"></div></div>
                                </div>
                                <div class="smart-gantt-timeline-row-footer">
                                    <div></div>
                                    <div><smart-button action="preview">${that.localize('preview')}</smart-button><smart-button action="submit" class="primary">${that.localize('submit')}</smart-button></div>
                                </div>
                            </div>`;
        }
        //Row default
        else if (!cell.innerHTML || !cell.firstElementChild.classList.contains('smart-gantt-timeline-row')) {
            cell.innerHTML = `<div class="smart-gantt-timeline-row">
                                <div class="smart-gantt-menu-button"></div>
                                <div class="smart-gantt-tickets">
                                    <div class="smart-gantt-ticket-container"></div>
                                </div>
                              </div>`;
        }

        const timelineRow = cell.firstElementChild,
            ticketsWrapper = timelineRow.querySelector('.smart-gantt-tickets'),
            ticketContainer = ticketsWrapper.firstElementChild;

        const { firstCellObj, lastCellObj } = that._getVisibleCellObjRange(),
            viewMin = firstCellObj.left,
            viewMax = lastCellObj.left,
            fragment = document.createDocumentFragment(),
            ticketElements = ticketContainer.children,
            scaleWidth = lastCellObj.width;
        let removedItemsCount = 0,
            left = -timeline.scrollView.scrollLeft,
            activeTicket;

        //Remove unnecessary tickets
        while (ticketElements.length > tickets.length) {
            ticketElements[0].remove();
        }

        for (let i = 0, max = tickets.length; i < max; i += 1) {
            const ticketObj = tickets[i];
            let ticketElement = ticketElements[i];

            that._validateTicketObj(ticketObj);

            if (!ticketObj || typeof ticketObj !== 'object') {
                if (ticketElement) {
                    ticketElement.classList.add('smart-visibility-hidden');
                }
                continue
            }

            if (ticketElement) {
                ticketElement.classList.remove('smart-visibility-hidden');
            }

            const minuteDiff = that._getMinuteDiff(ticketObj),
                width = ((minuteDiff) / 60) * scaleWidth;

            if (ticketObj.inProgress === 'Y' && !activeTicket) {
                activeTicket = ticketObj;
            }

            const viewWidth = viewMax - viewMin;
            //If the ticket ends before the first visible timeline cell or ends after the last, it will not be rendered
            if (left + width >= 0 && left <= viewWidth + width) {
                if (!ticketElement) {
                    ticketElement = document.createElement('div');
                    ticketElement.classList.add('smart-gantt-ticket');
                    fragment.appendChild(ticketElement);
                }

                ticketElement.setAttribute('start', ticketObj.start);
                ticketElement.setAttribute('end', ticketObj.end);
                ticketElement.classList.remove('smart-visibility-hidden');

                if (activeTicket === ticketObj) {
                    ticketElement.setAttribute('active', '');
                    ticketElement.setAttribute('progress', ticketObj.progress);
                }
                else {
                    ticketElement.removeAttribute('progress');
                    ticketElement.removeAttribute('active');
                }

                if (activeTicket !== ticketObj && ticketObj.duration < minuteDiff) {
                    ticketElement.setAttribute('delay', '');
                    ticketElement.style.setProperty('--delay', (Math.min(1, (1 - ticketObj.duration / minuteDiff)) * 100) + '%');
                }
                else {
                    ticketElement.removeAttribute('delay');
                    ticketElement.style.removeProperty('--delay');
                }

                ticketObj.completed === 'Y' ? ticketElement.setAttribute('completed', '') : ticketElement.removeAttribute('completed');
                ticketObj.type ? ticketElement.setAttribute('type', ticketObj.type) : ticketElement.removeAttribute('type');

                if (!ticketObj.type) {
                    ticketElement.classList.add('reorder');
                }

                that._setTicketTemplate(ticketElement, ticketObj);
                that._setTicketProggressBar(ticketElement, ticketObj);

                ticketElement.data = ticketObj;
                ticketElement.style.width = width + 'px';
            }
            else if (ticketElement) {
                ticketElement.classList.add('smart-visibility-hidden');
            }

            left += width;
        }

        //Increase the timelne range if the tickets overflow
        that._updateTimelineRange(left);

        if (fragment.childElementCount) {
            ticketContainer.appendChild(fragment);
        }

        const dropZoneWidth = timeline.scrollWidth - left;

        //Position and size for the tickets drop zone
        if (dropZoneWidth > 0) {
            ticketContainer.setAttribute('show-drop-zone', '');
            ticketContainer.style.setProperty('--smart-gantt-ticket-drop-zone-left', left + 'px');
            ticketContainer.style.setProperty('--smart-gantt-ticket-drop-zone-width', dropZoneWidth + 'px');
        }
        else {
            ticketContainer.removeAttribute('show-drop-zone');
            ticketContainer.style.removeProperty('--smart-gantt-ticket-drop-zone-left');
            ticketContainer.style.removeProperty('--smart-gantt-ticket-drop-zone-width');
        }

        //Sync the scrollLeft
        ticketsWrapper.scrollLeft = timeline.scrollView.scrollLeft;

        that._handleReorder(ticketContainer);
    }

    _createReorderFeedback(item) {
        const that = this;
        const feedback = document.createElement('div');

        feedback.className = item.className;
        feedback.classList.add('feedback', 'smart-visibility-hidden');
        feedback.innerHTML = item.innerHTML;

        if (item.hasAttribute('active')) {
            feedback.setAttribute('active', '');
        }

        if (item.hasAttribute('delay')) {
            feedback.setAttribute('delay', '');
        }

        if (item.hasAttribute('completed')) {
            feedback.setAttribute('completed', '');
        }

        if (that.theme) {
            feedback.setAttribute('theme', that.theme);
        }

        feedback.style.width = item.offsetWidth + 'px';
        feedback.style.height = item.offsetHeight + 'px';

        document.body.appendChild(feedback);
        that._ticketDragDetails.feedback = feedback;
    }

    /**
* Handles tickets reorder in a single row.
*/
    _handleReorder(ticketContainer) {
        const that = this;

        ticketContainer.onpointerdown = (event) => {
            const ticketReoderItems = [...ticketContainer.querySelectorAll('.smart-gantt-ticket')].filter((item) => {
                return !item.hasAttribute('type') && !item.classList.contains('smart-visibility-hidden');
            });;

            const clickedItem = (that.shadowRoot || that.isInShadowDOM ? event.originalEvent.composedPath()[0] : event.target).closest('.smart-gantt-ticket');

            if (!clickedItem || (clickedItem && clickedItem.hasAttribute('type'))) {
                return;
            }

            that._ticketDragDetails = {
                container: ticketContainer,
                items: ticketReoderItems,
                item: clickedItem,
                x: event.pageX,
                y: event.pageY,
                coordinates: that._getItemCoordinates(ticketReoderItems)
            };

            that._updateReorderIndex(event);
            that._ticketDragDetails.reorderStartIndex = that._ticketDragDetails.index;
        }
    }

    /**
* Updates the index of the reorder details object. The index is set to the item's index under the pointer.
*/
    _updateReorderIndex(event) {
        const that = this;

        const coordinates = that._ticketDragDetails.coordinates;
        const items = that._ticketDragDetails.items;

        that._ticketDragDetails.index = -1;
        for (let i = 0; i < coordinates.length; i++) {
            const currentTicketCoordinate = coordinates[i];
            const left = that._timeline.scrollView.scrollLeft;
            const offsetLeft = currentTicketCoordinate.offset.left - left;
            const offsetTop = currentTicketCoordinate.offset.top;

            if (event.pageY >= offsetTop - 5 && event.pageY <= offsetTop + currentTicketCoordinate.height + 5) {
                if (event.pageX >= offsetLeft && event.pageX < offsetLeft + currentTicketCoordinate.width) {
                    that._ticketDragDetails.index = i;
                }

                if (i === coordinates.length - 1 && (event.pageX >= offsetLeft + currentTicketCoordinate.width / 2)) {
                    that._ticketDragDetails.index = i + 1;
                }
            }
        }
    }

    /**
* Clears the placeholder.
*/
    _clearReorderPlaceholder() {
        const that = this;

        const items = that._ticketDragDetails.items;

        for (let i = 0; i < items.length; i++) {
            items[i].classList.remove('reorder-placeholder');
            items[i].classList.remove('reorder-placeholder-after');
        }
    }

    _documentMoveHandler(event) {
        const that = this;

        super._documentMoveHandler(event);

        that._handleTicketsReoderMove(event);
    }

    _handleAutoScrollOnTicketDrag(event) {
        const that = this;
        const buttons = that.querySelectorAll('.smart-timeline-button');
        const autoScrollStartDistance = 50;

        const autoScrollLeft = () => {
            const left = that.offset(buttons[0]).left;

            if (event.pageX >= left - autoScrollStartDistance && event.pageX <= left + autoScrollStartDistance) {
                return true;
            }

            return false;
        }

        const autoScrollRight = () => {
            const right = that.offset(buttons[1]).left;

            if (event.pageX >= right - autoScrollStartDistance && event.pageX <= right + autoScrollStartDistance) {
                return true;
            }

            return false;
        }
        if (that._autoScrollInterval) {
            clearInterval(that._autoScrollInterval);
        }

        const refreshDraggedItems = () => {
            const ticketReoderItems = [...that._ticketDragDetails.container.querySelectorAll('.smart-gantt-ticket')].filter((item) => {
                return !item.hasAttribute('type') && !item.classList.contains('smart-visibility-hidden');
            });;

            that._ticketDragDetails.items = ticketReoderItems;
            that._ticketDragDetails.coordinates = that._getItemCoordinates(that._ticketDragDetails.items);
        }

        if (autoScrollLeft() || autoScrollRight()) {
            that._autoScrollInterval = setInterval(() => {
                if (autoScrollLeft()) {
                    that._handleTimelineButtonClick(buttons[0]);
                    refreshDraggedItems();
                }
                else if (autoScrollRight()) {
                    that._handleTimelineButtonClick(buttons[1]);
                    refreshDraggedItems();
                }
            }, 50);
        }
    }

    /**
* Starts a reorder and updates the position of the feedback element.
*/
    _handleTicketsReoderMove(event) {
        const that = this;

        if (!that._ticketDragDetails) {
            return;
        }

        const x = that._ticketDragDetails.x;
        const y = that._ticketDragDetails.y;

        if (that._ticketDragDetails.reorderStarted) {
            const reorderItem = that._ticketDragDetails.item;
            const items = that._ticketDragDetails.items;

            that._updateReorderIndex(event);
            that._clearReorderPlaceholder();
            that._handleAutoScrollOnTicketDrag(event);

            that._ticketDragDetails.feedback.style.left = 5 + event.pageX + 'px';
            that._ticketDragDetails.feedback.style.top = 5 + event.pageY + 'px';
            that._ticketDragDetails.feedback.classList.remove('smart-visibility-hidden');

            const index = that._ticketDragDetails.index;

            if (index === -1) {
                that._ticketDragDetails.dropOverItem = null;
                that._ticketDragDetails.feedback.classList.add('no-drop');
                return;
            }

            that._ticketDragDetails.feedback.classList.add('can-drop');

            const currentItem = items[that._ticketDragDetails.index];

            if (currentItem) {
                currentItem.classList.add('reorder-placeholder');
                that._ticketDragDetails.dropOverItem = currentItem;
            }
            else {
                const lastItem = items[that._ticketDragDetails.index - 1];

                if (lastItem) {
                    lastItem.classList.add('reorder-placeholder');
                    lastItem.classList.add('reorder-placeholder-after');
                    that._ticketDragDetails.dropOverItem = lastItem;
                }
            }
        }

        if (!(Math.abs(event.pageX - x) >= 3 || Math.abs(event.pageY - y) >= 3)) {
            return;
        }

        if (that._ticketDragDetails.reorderStarted) {
            return;
        }

        that._ticketDragDetails.reorderStarted = true;
        const row = event.originalEvent.target.closest('tbody tr[row-id]');
        const id = row.getAttribute('row-id');
        const reorderItem = that._ticketDragDetails.item;

        that._ticketDragDetails.rowId = id;
        that._ticketDragDetails.row = row;

        reorderItem.classList.add('reorder-start');
        that._createReorderFeedback(reorderItem);
        that.$.fireEvent('ticketDragStart', {
            data: that._ticketDragDetails.item.data,
            id: id
        });
    }

    _offsetTop(element) {
        const that = this;

        if (!element) {
            return 0;
        }

        return element.offsetTop + that._offsetTop(element.offsetParent);
    }

    _offsetLeft(element) {
        const that = this;

        if (!element) {
            return 0;
        }

        return element.offsetLeft + that._offsetLeft(element.offsetParent);
    }

    offset(element) {
        return { left: this._offsetLeft(element), top: this._offsetTop(element) }
    }

    _documentUpHandler(event) {
        const that = this;

        super._documentUpHandler(event);

        that._handleTicketsDrop(event);
    }

    _tableContainerKeyupHandler(event) {
        const that = this;

        if (event.key === 'Escape') {
            that._cancelTicketReorder();
        }
    }

    _cancelTicketReorder() {
        const that = this;
        if (that._ticketDragDetails && that._ticketDragDetails.reorderStarted) {
            that.$.fireEvent('ticketDragCancel', {
                data: that._ticketDragDetails.item.data
            });

            that._clearReorderResources();
        }
    }

    _clearReorderResources() {
        const that = this;

        if (that._autoScrollInterval) {
            clearInterval(that._autoScrollInterval);
        }
        that._ticketDragDetails.item.classList.remove('reorder-start');
        that._clearReorderPlaceholder();
        that._ticketDragDetails.feedback.remove();
        that._ticketDragDetails = null;

    }
    /**
     * Ends the tickets reorder operation.
     */
    _handleTicketsDrop(event) {
        const that = this;

        if (!that._ticketDragDetails || (that._ticketDragDetails && !that._ticketDragDetails.reorderStarted)) {
            that._ticketDragDetails = null;
            return;
        }

        const row = event.originalEvent.target.closest('tbody tr[row-id]');

        if (row) {
            const id = row.getAttribute('row-id');
            const item = that._ticketDragDetails.item;
            const ticketData = item.data;
            const tickets = row.data['timeline'];

            that.$.fireEvent('ticketReorder', {
                data: that._ticketDragDetails.item.data,
                endData: that._ticketDragDetails.dropOverItem ? that._ticketDragDetails.dropOverItem.data : null,
                id: id
            });

            // perform same row tickets reorder.
            if (that._ticketDragDetails.rowId === id) {
                const dropOverItem = that._ticketDragDetails.dropOverItem;
                const dropOverTicketData = dropOverItem.data;
                if (dropOverItem) {
                    const itemIndex = tickets.indexOf(ticketData);
                    const newIndex = tickets.indexOf(dropOverTicketData);

                    const move = (array, oldIndex, newIndex) => {
                        if (newIndex >= array.length) {
                            return;
                        }

                        newIndex = Math.max(1, newIndex);

                        array.splice(newIndex, 0, array.splice(oldIndex, 1)[0]);
                        that._recycleRowTickets();
                    };

                    if (that._ticketDragDetails.dropOverItem.classList.contains('reorder-placeholder-after')) {
                        move(tickets, itemIndex, newIndex);
                    }
                    else {
                        move(tickets, itemIndex, newIndex - 1);
                    }
                }
            }
            else {
                const ticketContainer = event.originalEvent.target.closest('.smart-gantt-ticket-container');

                if (ticketContainer) {
                    let isDropZoneDrop = false;

                    const ticketContainerOffset = that.offset(ticketContainer);
                    const isDropInTimeline = event.pageX >= ticketContainerOffset.left && event.pageX <= ticketContainerOffset.left + ticketContainer.offsetWidth;
                    const ticketElements = [...ticketContainer.querySelectorAll('.smart-gantt-ticket')];
                    const hiddenTicketElements = ticketElements.filter((ticket) => {
                        return ticket.classList.contains('.smart-visibility-hidden');
                    });

                    if (isDropInTimeline) {
                        if (ticketElements.length === hiddenTicketElements.length) {
                            isDropZoneDrop = true;
                        }

                        const lastTicket = ticketElements[ticketElements.length - 1];

                        if (lastTicket) {
                            const timelineScrollPosition = that.getTimelineScrollPosition();

                            if (!lastTicket.classList.contains('.smart-visibility-hidden') && that.offset(lastTicket).left - timelineScrollPosition <= event.pageX) {
                                isDropZoneDrop = true;
                            }
                        }
                    }

                    if (isDropZoneDrop) {
                        const tickets = that._ticketDragDetails.row.data['timeline'];
                        const dropRowTickets = row.data['timeline'];

                        const itemIndex = tickets.indexOf(ticketData);
                        tickets.splice(itemIndex, 1);
                        dropRowTickets.splice(dropRowTickets.length - 2, 0, ticketData);
                        that._recycleRowTickets();
                    }
                }
            }
        }

        that.$.fireEvent('ticketDragEnd', {
            data: that._ticketDragDetails.item.data
        });
        that._clearReorderResources();
    }

    /**
    * Gets item coordinates.
    */
    _getItemCoordinates(items) {
        const that = this,
            coordinates = [];

        items.forEach(function (item) {
            const rect = item.getBoundingClientRect();

            coordinates.push({
                offset: that.offset(item),
                left: rect.left,
                top: rect.top,
                right: rect.right,
                bottom: rect.bottom,
                width: rect.width,
                height: rect.height
            });
        });

        return coordinates;
    }

    _getMinuteDiff(ticketObj) {
        const that = this;
        let [startHour, startMinute] = ticketObj.start.split(':').map(i => parseInt(i)),
            [endHour, endMinute] = ticketObj.end.split(':').map(i => parseInt(i));

        while (endHour < startHour) {
            endHour = (endHour + 1) % 24;
            startHour = (startHour + 1) % 24;
        }

        //Min time of a ticket is 15min
        return Math.max(15, (endHour * 60 + endMinute) - (startHour * 60 + startMinute))
    }

    /**
     * Updates the Timeline if there are tickets that overflow the default time range
     * @param {number} ticketEnd - the right position of the last ticket element
     */
    _updateTimelineRange(ticketEnd) {
        const that = this,
            timeline = that._timeline,
            cellsObjs = timeline.cellObjs,
            lastCellObj = cellsObjs[cellsObjs.length - 1],
            scaleWidth = lastCellObj ? lastCellObj.width : 50, // TODO: Create a CSS var or JS property for the width of the scale cells
            scaleEnd = lastCellObj.left + scaleWidth;

        if (scaleEnd < ticketEnd) {
            const cellsCount = Math.ceil((ticketEnd - scaleEnd) / scaleWidth);
            let left = scaleEnd, hour = lastCellObj.hour;

            for (let i = 0; i < cellsCount; i += 1) {
                hour = (hour + 1) % 24;

                cellsObjs.push({
                    hour: hour,
                    width: scaleWidth,
                    left: left
                });

                left += scaleWidth;
            }

            if (cellsCount) {
                timeline.scrollWidth = Math.max(cellsObjs.length * scaleWidth, timeline.scrollView.offsetWidth);
                that.$.tableContainer.style.setProperty('--smart-gantt-timeline-width', timeline.scrollWidth + 'px');
            }
        }

        that._recycleTimeline();
        that._refreshScrollButtons();
    }

    /**
     * Set ticket progress bar
     * @param {HTMLElement} ticketElement - the ticket element
     * @param {object} ticketObj - the ticket object
     * @returns 
     */
    _setTicketProggressBar(ticketElement, ticketObj) {
        const that = this,
            progress = ticketObj.progress,
            thumb = ticketElement.firstElementChild;

        if (!thumb || !thumb.classList.contains('smart-gantt-ticket-progress-thumb') || !progress || progress === 100 || !ticketElement.hasAttribute('active')) {
            return
        }

        thumb.style[that.rightToLeft ? 'right' : 'left'] = (Math.min(100, Math.max(0, progress)) || 0) + '%';
    }

    /**
     * Sets the template of a ticket
     */
    _setTicketTemplate(target, ticketObj) {
        const that = this,
            label = ticketObj.label || '',
            labelElement = target.children[1];
        let template = ticketObj.template;

        if (!labelElement || !labelElement.classList.contains('smart-gantt-ticket-task-label')) {
            target.innerHTML = !target.hasAttribute('active') || target.hasAttribute('completed') ? `</div><div class="smart-gantt-ticket-content">${label}</div>` :
                `<div class="smart-gantt-ticket-progress-thumb"></div><div class="smart-gantt-ticket-content">${label}</div>`;

            if (!template) {
                return
            }
        }
        else if (!template) {
            labelElement.textContent = label;
            return
        }

        //Handle groupTemplate
        if (typeof template === 'function') {
            template.call(that, labelElement, ticketObj);
            delete that._ticketTemplateBindings;
            return;
        }
        else if (typeof template === 'string') {
            template = document.getElementById(template);
        }

        if (!('content' in document.createElement('template'))) {
            that.error(that.localize('htmlTemplateNotSuported', { elementType: that.nodeName.toLowerCase() }));
            delete that._ticketTemplateBindings;
            return;
        }

        if (template === null || !('content' in template)) {
            that.error(that.localize('invalidTicketTemplate', { elementType: that.nodeName.toLowerCase(), property: 'template' }));
            delete that._ticketTemplateBindings;
            return;
        }

        if (!that._ticketTemplateBindings) {
            that._ticketTemplateBindings = {};
        }

        labelElement.innerHTML = '';

        let templateBindings = that._ticketTemplateBindings, bindings;
        const clone = document.importNode(template.content, true),
            nodes = clone.childNodes;

        if (templateBindings) {
            bindings = templateBindings;
        }

        //Handles template bindings
        for (let i = 0; i < nodes.length; i++) {
            let nodeHTML = nodes[i].outerHTML;

            if (nodeHTML) {
                if (!templateBindings) {
                    const regex = /{{\s*\w*\s*}}/gm;

                    //Replace bindings
                    if (regex.test(nodeHTML)) {
                        const matches = nodeHTML.match(regex);

                        if (!bindings) {
                            bindings = [];
                        }

                        if (matches) {
                            matches.forEach(m => {
                                if (!bindings.includes(m)) {
                                    bindings.push(m);
                                }
                            });
                        }
                    }
                }

                if (bindings) {
                    for (let i = 0; i < bindings.length; i++) {
                        const match = bindings[i];
                        nodeHTML = nodeHTML.replace(match, ticketObj[match.replace(/{{|}}/gm, '').trim()] || '');
                    }
                }

                labelElement.innerHTML += nodeHTML;
            }
        }

        if (bindings) {
            that._ticketTemplateBindings = bindings;
        }
    }

    /**
     * Validates the properties of the ticketObj
     */
    _validateTicketObj(ticketObj) {
        if (typeof ticketObj !== 'object') {
            return
        }

        const ticketClass = ticketObj.class,
            ticketType = ticketObj.type,
            ticketDuration = ticketObj.duration,
            ticketProgress = ticketObj.progress,
            ticketDelay = ticketObj.delay,
            ticketTypes = ['ticket', 'preRoute', 'lunchTime', 'returnYard', 'returnToYard', 'postRoute'];

        // ticketObj.dateStart = new Date(ticketObj.dateStart);

        // if(ticketType) {
        //     ticketObj.type = ticketTypes.indexOf(ticketType) < 0 ? '' : ticketType
        // }

        // if (ticketClass) {
        //     ticketObj.class = typeof ticketClass !== 'string' ? '' : ticketClass;
        // }

        if (ticketDuration !== undefined) {
            ticketObj.duration = typeof ticketDuration === 'number' && !isNaN(ticketDuration) ? ticketDuration : 0;
        }

        if (ticketProgress !== undefined) {
            ticketObj.progress = typeof ticketProgress === 'number' && !isNaN(ticketProgress) ? ticketProgress : (parseFloat(ticketProgress.toFixed(2)));
        }

        if (ticketDelay) {
            ticketObj.delay = typeof ticketDelay === 'number' && !isNaN(ticketDelay) ? parseInt(ticketDelay) : 0;
        }
    }

    /**
     * Sets the row details
     * @param {Proxy} rowObject - contains row details
     * @param {HTMLElement} rowDetail - the Table row that contains the details
     */
    _setRowDetail(rowObject, rowDetail) {
        const that = this,
            table = rowDetail.querySelector('smart-table');

        if (!table) {
            return
        }

        const tickets = rowObject.data['timeline'].filter(i => !i.type),
            columns = that.columns,
            rowDetailColumns = ['options', 'name', 'label', 'completed', 'transmission', 'routeOrderData', 'empty', 'empty', 'unassign'];
        let detailColumns = [], nextColumnWidth;

        for (let i = 0, max = columns.length; i < max; i += 1) {
            const column = columns[i],
                dataField = rowDetailColumns[i],
                newColumn = {};

            // if (column.width) {
            newColumn.label = '';

            if (i > 0) {
                newColumn.width = column.width !== undefined ? column.width : ((dataField === 'name' ? 2 : 1) * column.headerCellElement.offsetWidth);
            }

            newColumn.dataField = dataField;

            switch (dataField) {
                case 'unassign':
                    newColumn.formatFunction = (settings) => settings.value = that.localize('unassign') || '';
                    break;
                case 'completed':
                    newColumn.formatFunction = (settings) => {
                        const completed = settings.data.completed,
                            cell = settings.cell;

                        settings.value = '';

                        completed === 'Y' ? cell.setAttribute('completed', '') : cell.removeAttribute('completed');
                    };
                    break;
                case 'timeline':
                    // newColumn.formatFunction = (settings) => {

                    // }
                    break;
            }

            detailColumns.push(newColumn);
        }

        table.dataSource = tickets;
        table.columns = detailColumns;
        table.rightToLeft = that.rightToLeft;
        table.theme = that.theme;

        if (table.wait === false && table.hasAttribute('wait')) {
            table.wait = true;
        }
        table.wait = false;
    }

    /**
    * Sets cell content.
    */
    _setCellContent(cell, content, isTreeCell, rowId) {
        const that = this;

        if (content instanceof HTMLElement !== true) {
            if (isTreeCell) {
                cell.innerHTML = content = `<div>
                                                <div class="hierarchy-arrow smart-arrow smart-arrow-down" role="button" aria-label="Toggle row"></div>
                                                <div>${content}</div>
                                            </div>`;
            }
            else if (typeof (content) === 'string' && content.indexOf('>') >= 0) {
                if (/<.+?>/.test(content) || /&\S+;/.test(content)) {
                    if (cell._content === content) {
                        return;
                    }

                    cell.innerHTML = content;
                    cell._content = content;
                }
                else {
                    if (cell.textContent !== content) {
                        cell.textContent = content;
                    }
                }
            }
            else if (cell.getAttribute('data-field') === 'timeline') {
                //Handle Tickets
                if (!Array.isArray(content)) {
                    content = [content];
                }

                that._refreshTickets(content, cell, rowId);
            }
            else {
                if (cell.textContent !== content) {
                    if (typeof content === 'object' && cell.getAttribute('data-field') === 'progress') {
                        cell.textContent = `${content.completed}/${content.total}`;
                    }
                    else {
                        cell.textContent = content;
                    }
                }
            }

            return;
        }

        while (cell.childNodes.length > 0) {
            cell.childNodes[0].remove();
        }

        if (isTreeCell) {
            const container = document.createElement('div');

            container.innerHTML = '<div class="hierarchy-arrow smart-arrow smart-arrow-down" role="button" aria-label="Toggle row"></div><div></div>';
            container.children[1].appendChild(content);
            cell.appendChild(container);
        }
        else {
            cell.appendChild(content);
        }
    }

    /**
     * Returns the start/end timeline cell objects for the ticket depending on it's start/end time
     * @param {*} ticketStart 
     * @param {*} ticketEnd 
     */
    _getTicketCellObjRange(ticketStart, ticketEnd) {
        const that = this,
            timeline = that._timeline;
        let cellStart, cellEnd;

        if (!timeline || !timeline.cellObjs || !timeline.cellObjs.length) {
            return { cellStart: cellStart, cellEnd: cellEnd }
        }

        const cellObjs = timeline.cellObjs,
            oneHour = 60 * 60 * 1000; // 1 hour minus 1 milisecond

        for (let i = 0, max = cellObjs.length; i < max; i += 1) {
            const cellObj = cellObjs[i],
                cellObjStart = cellObj.time,
                cellObjEnd = cellObjStart + oneHour;

            if (!cellStart && ticketStart >= cellObjStart && ticketStart < cellObjEnd) {
                cellStart = cellObj
            }

            if (!cellEnd && ticketEnd >= cellObjStart && ticketEnd <= cellObjEnd) {
                cellEnd = cellObj
            }

            if (cellStart && cellEnd) {
                break;
            }
        }

        if (cellStart && !cellEnd) {
            cellEnd = cellObjs[cellObjs.length - 1];
        }

        return { cellStart: cellStart, cellEnd: cellEnd }
    }

    /**
     * Retursn the first visible timeline cell obj
     * @returns 
     */
    _getVisibleCellObjRange() {
        const that = this,
            timeline = that._timeline;
        let firstCellObj, lastCellObj;

        if (!timeline) {
            return { firstCellObj: firstCellObj, lastCellObj: lastCellObj }
        }

        const cellObjs = timeline.cellObjs;

        if (!cellObjs) {
            return { firstCellObj: firstCellObj, lastCellObj: lastCellObj }
        }

        const [offset, size] = ['left', 'width'],
            scrollWidth = timeline.scrollView.offsetWidth,
            scrollSize = Math.abs(timeline.scrollView.scrollLeft);

        //Find the first cell in view depending on the current scrollLeft
        for (let i = 0; i < cellObjs.length; i++) {
            const cellObj = cellObjs[i];

            if (!firstCellObj && cellObj[offset] + cellObj[size] >= scrollSize) {
                firstCellObj = cellObj;
            }

            if (cellObj[offset] + cellObj[size] >= scrollSize + scrollWidth) {
                lastCellObj = cellObj;
            }

            if (firstCellObj && lastCellObj) {
                break;
            }

            // lastCellObj = cellObj;
        }

        if (!lastCellObj) {
            lastCellObj = cellObjs[cellObjs.length - 1];
        }

        if (!firstCellObj && lastCellObj) {
            firstCellObj = scrollSize > lastCellObj[offset] ? lastCellObj : cellObjs[0];
        }

        return { firstCellObj: firstCellObj, lastCellObj: lastCellObj }
    }

    /**
     * Returns the number of visible cells for the scale to be displayed inside the Timeline
     */
    _getVisibleCellsCount(firstCellObj) {
        const that = this,
            timeline = that._timeline;

        if (!timeline || !timeline.element) {
            return
        }

        const cellObjs = timeline.cellObjs;

        if (!cellObjs) {
            return 0
        }

        if (!firstCellObj) {
            ({ firstCellObj } = that._getVisibleCellObjRange());
        }

        const [offset, size] = ['left', 'width'],
            scrollView = timeline.scrollView,
            scrollLeft = scrollView.scrollLeft,
            scrollSize = Math.abs(scrollLeft || 0);

        if (!firstCellObj) {
            return 0
        }

        //Calculate the number of scale cells needed
        const viewSize = timeline.scrollView.offsetWidth,
            firstCellOffset = 1 - (firstCellObj[offset] + firstCellObj[size] - scrollSize) / cellObjs[0][size];

        if (isNaN(firstCellOffset) || !isFinite(firstCellOffset)) {
            return 0
        }

        //Create an additional cell if scrollMax is reacted
        return Math.min(cellObjs.length, Math.ceil(parseFloat(((viewSize / firstCellObj[size]) + firstCellOffset).toFixed(2))))
    }

    _getTimelineScaleCount() {
        const that = this;

        //TODO: Go through the dataSource to find the start and end dates
        //NOTE: Min/max scale time will be determined by properties 

        return 24
    }

    /**
    * Creates Gantt header
    */
    _createTableHeader() {
        const that = this,
            selection = that.selection,
            thead = document.createElement('thead'),
            columns = that._columns.filter(col => col.visible),
            levels = that._getColumnHeaderStructure(),
            windowWidth = window.innerWidth,
            hiddenColumnIndexes = {};

        for (let l = 0; l < levels; l++) {
            const tr = document.createElement('tr'),
                fragment = document.createDocumentFragment();

            if (selection) {
                if (l < levels - 1) {
                    tr.innerHTML = '<th class="empty"></th>';
                }
                else {
                    tr.innerHTML = '<th class="smart-table-select-all freeze-near" scope="col" aria-colindex="1"><div role="checkbox" aria-checked="false" aria-label="Toggle selection of all rows"></div></th>';
                }
            }

            //Detail column
            if (that.rowDetailTemplate) {
                tr.innerHTML += '<th class="smart-table-detail-col freeze-near" aria-colindex="1"></th>';
            }

            if (l < levels - 1) {
                tr.classList.add('column-groups-header');
            }

            for (let i = 0; i < columns.length; i++) {
                const column = columns[i];

                if (l === 0) {
                    // one-time column settings
                    if (column.width && isNaN(parseFloat(column.width))) {
                        delete column.width;
                    }

                    if (column.responsivePriority) {
                        if (typeof column.responsivePriority !== 'number') {
                            column.responsivePriority = 1;
                        }

                        column.responsivePriority = Math.max(Math.min(Math.round(column.responsivePriority), 5), 1);

                        if (levels > 1) {
                            hiddenColumnIndexes[i] = that._isColumnHidden(column.responsivePriority, windowWidth);
                        }
                    }
                }

                if (l < levels - 1) {
                    if (hiddenColumnIndexes[i]) {
                        continue;
                    }

                    const column = columns[i],
                        label = column.headerStructure[l];

                    if (i > 0) {
                        let position = i - 1,
                            previousColumn = columns[position];

                        while (previousColumn) {
                            if (that._isColumnHidden(previousColumn.responsivePriority, windowWidth)) {
                                position--;
                                previousColumn = columns[position];
                            }
                            else {
                                break;
                            }
                        }

                        if (previousColumn) {
                            const previousLabel = previousColumn.headerStructure[l];

                            if (label === previousLabel &&
                                that._areColumnsWithSameFreezeSettings(column, previousColumn)) {
                                continue;
                            }
                        }
                    }

                    const th = document.createElement('th');

                    th.includes = [column];

                    if (i < columns.length - 1) {
                        let colSpan = 1,
                            position = i + 1,
                            nextColumn = columns[position];

                        while (nextColumn) {
                            if (!that._isColumnHidden(nextColumn.responsivePriority, windowWidth)) {
                                const nextLabel = nextColumn.headerStructure[l];

                                if (label === nextLabel &&
                                    that._areColumnsWithSameFreezeSettings(column, nextColumn)) {
                                    colSpan++;
                                    th.includes.push(nextColumn);
                                }
                                else {
                                    break;
                                }
                            }

                            position++;
                            nextColumn = columns[position];
                        }

                        th.colSpan = colSpan;
                    }

                    if (l === 0 || column.headerStructure[l - 1] !== label) {
                        th.textContent = label;
                    }

                    th.includes.forEach(col => col.thHierarchy[l] = th);

                    th.classList.add('column-header');

                    if (column.freeze) {
                        th.classList.add('freeze-' + (column.freeze === 'far' ? 'far' : 'near'));
                    }

                    fragment.appendChild(th);
                }
                else {
                    that._setupColumnHeaderCell(selection, fragment, i, column);
                }
            }

            tr.appendChild(fragment);
            thead.appendChild(tr);
        }

        that.$.tableContainer.appendChild(thead);

        //Create the timeline header
        that._createHeaderTimeline();

        if (that.columnSizeMode === 'default') {
            for (let i = 0; i < that.columns.length; i++) {
                const column = that.columns[i];

                if (!column.width) {
                    const th = thead.querySelector('[data-field="' + column.dataField + '"]');

                    if (th && th.offsetWidth < 5) {
                        th.style.width = parseInt(that.columnMinWidth) + 'px';
                    }
                }
            }

            if (selection) {
                const visibleColumns = that.columns.filter((column) => {
                    return column.visible;
                });

                if (visibleColumns.length === 1) {
                    const column = visibleColumns[0];
                    const th = thead.querySelector('[data-field="' + column.dataField + '"]');

                    if (th) {
                        th.style.width = '';
                    }
                }
            }
        }
    }

    /**
     * Handles menu item click
     * @param {Event} event 
     */
    _handleMenuItemClick(rowId, event) {
        const that = this,
            detail = event.detail;

        if (detail.value === 'editRoute') {
            that.beginRowEdit(rowId);
        }
    }

    /**
     * Handles Timeline scroll button click
     * @param {HTMLElement} button - a Smart.RepeatButton
     * @param {number} step - the scroll step
     * @returns 
     */
    _handleTimelineButtonClick(button, step = 20) {
        const that = this,
            timeline = that._timeline;
        let scroll;

        if (!button || !timeline) {
            return;
        }

        if (button instanceof Smart.RepeatButton) {
            scroll = step * (button.hasAttribute('prev') ? -1 : 1);
        }

        that.scrollTimelineTo(scroll);
    }

    scrollTimelineTo(value) {
        const that = this;

        that._timeline.scrollView.scrollLeft += that.rightToLeft ? (value * -1) : value;

        that._recycleTimeline();
        that._refreshScrollButtons();
        that._recycleRowTickets();
    }

    getTimelineScrollPosition() {
        const that = this;

        return that._timeline.scrollView.scrollLeft;
    }
    // /**
    //  * Handles ticket click event
    //  * @param { HTMLElement } ticket - the ticket element that was clicked
    //  */
    // _handleTicketClick(ticket) {
    //     const that = this;

    //     if (!ticket || !ticket.classList.contains('smart-gantt-ticket')) {
    //         return
    //     }

    //     // ticket.classList.toggle('active');
    // }

    /**
     * Refreshes the Timeline scroll buttons
     * @returns 
     */
    _refreshScrollButtons() {
        const that = this,
            timeline = that._timeline;

        if (!timeline) {
            return;
        }

        const scrollTarget = that._timeline.scrollView,
            viewButtons = that._timeline.buttons;

        if (!scrollTarget || !viewButtons) {
            return;
        }

        const scroll = scrollTarget.scrollLeft,
            prevButton = viewButtons.prev,
            nextButton = viewButtons.next;

        //NOTE: Bug with RTL. ScrollWidth can become +/- 1px
        if (prevButton) {
            prevButton.disabled = !scroll;
        }

        if (nextButton) {
            nextButton.disabled = scrollTarget.offsetWidth + Math.abs(scroll) === scrollTarget.scrollWidth;
        }
    }

    /**
     * Processes and rendereds the Details expand/collapse button of a row
     * @param {*} data 
     * @param {*} rowElement 
     * @param {*} hierarchical 
     * @param {*} update 
     */
    _processAndRenderDetails(data, rowElement, hierarchical, update) {
        const that = this;
        // id = data.$.id,

        // if (hierarchical && that.selectionByHierarchy) {
        //     const selectedIds = that._selectedIds,
        //         selectedIdIndex = selectedIds.indexOf(id);

        //     if (state === true && selectedIdIndex === -1) {
        //         that._selectedIds.push(id);
        //     }
        //     else if (!state && selectedIdIndex !== -1) {
        //         that._selectedIds.splice(selectedIdIndex, 1);
        //     }
        // }

        if (!that.rowDetailTemplate) {
            return;
        }

        let detailButtonTd, checkbox;
        const state = that._expandedRowDetailIds.indexOf(data.$.id) > -1;

        if (update) {
            detailButtonTd = rowElement.firstElementChild;
            checkbox = detailButtonTd.firstElementChild;
        }
        else {
            detailButtonTd = document.createElement('td');
            checkbox = document.createElement('div');
        }

        checkbox.className = 'detail-checkbox';
        checkbox.setAttribute('role', 'checkbox');
        checkbox.setAttribute('aria-checked', state);
        checkbox.setAttribute('aria-label', 'Toggle row selection');
        detailButtonTd.className = `smart-table-detail-row freeze-near${state ? ' selected' : ''}`;
        detailButtonTd.setAttribute('aria-colindex', 1);

        if (!update) {
            detailButtonTd.appendChild(checkbox);
            rowElement.appendChild(detailButtonTd);
        }

        rowElement.setAttribute('aria-expanded', state);
    }

    /**
     * Sets up a column's header cell
     */
    _setupColumnHeaderCell(selection, fragment, i, column) {
        const that = this,
            th = document.createElement('th');

        th.setAttribute('aria-colindex', i + 1 + (+selection));

        if (column.width) {
            th.style.width = that._getColumnWidth(column.width, true);
        }
        else if (column._manualWidth) {
            if (column._isResponsive) {
                th.style.width = column._manualPercentageWidth + '%';
            }
            else {
                th.style.width = column._manualWidth + 'px';
            }
        }

        if (that._columns.length === 1 || !column.allowResize) {
            th.setAttribute('locked', '');
        }

        if (column.freeze) {
            th.classList.add('freeze-' + (column.freeze === 'far' ? 'far' : 'near'));
        }

        if (column.responsivePriority) {
            th.classList.add('priority-' + column.responsivePriority);
        }

        th.setAttribute('data-field', column.dataField);

        if (column.dataField === 'timeline') {
            let timeline = that._timeline;

            th.innerHTML = `<div class="smart-gantt-timeline-column">
                                <smart-repeat-button class="smart-timeline-button" prev></smart-repeat-button>
                                    <div class="smart-timeline-container">
                                        <div class="smart-timeline"></div>
                                    </div>
                                <smart-repeat-button class="smart-timeline-button" next></smart-repeat-button>
                             </div>`;

            //Store the timeline element
            if (!timeline) {
                timeline = that._timeline = {};
            }

            if (!timeline.element) {
                //NOTE: QuerySelectorAll returns the elements in their DOM order
                const elements = th.querySelectorAll('.smart-timeline-button[prev], .smart-timeline, .smart-timeline-button[next]');

                timeline.element = elements[1];
                timeline.scrollView = elements[1].parentElement;
                timeline.buttons = { prev: elements[0], next: elements[2] };
            }
        }
        else {
            th.innerHTML = `<div class="wrapper" role="presentation"><div class="label">${column.label}</div></div>`;
        }

        if (that.onColumnRender) {
            that.onColumnRender(column.dataField, th);
        }

        fragment.appendChild(th);
    }

    /**
    * tableContainer click handler.
    */
    _tableContainerClickHandler(event) {
        const that = this,
            target = that.isInShadowDOM ? event.composedPath()[0] : event.target;

        if (!target.closest) {
            return
        }

        const timelineButton = target.closest('.smart-timeline-button');

        if (timelineButton) {
            that._handleTimelineButtonClick(timelineButton);
            return
        }

        const rowMenuButton = target.closest('.smart-gantt-menu-button');

        if (rowMenuButton) {
            that._handleRowMenu(rowMenuButton);
            return
        }

        if (that._editing || !that.$.tableContainer.contains(target)) {
            return;
        }

        const row = target.closest('tbody tr[row-id], tfoot tr.grand-total');

        if (row) {
            const cell = target.closest('td');

            if (target.classList.contains('hierarchy-arrow')) {
                const animation = that.animation;

                if (that.virtualization) {
                    that._animation = 'none';
                }

                that._hierarchyArrowClickHandler(row, cell);

                if (that.virtualization) {
                    that._animation = animation;
                }
                return;
            }

            const id = row.getAttribute('row-id');

            let rowData;

            if (id) {
                rowData = that.rowById ? that.rowById[id].data :
                    that.rows.dataItemById[id];
            }

            const parsedId = rowData ? rowData.$.id : undefined,
                dataField = cell ? cell.getAttribute('data-field') : undefined;

            if (!rowData && row.classList.contains('grand-total')) {
                rowData = row.data;
            }

            if (dataField !== undefined) {
                that.$.fireEvent('cellClick', { id: id, row: rowData, dataField: dataField, value: rowData[dataField], originalEvent: event });
            }

            that._toggleSelection(row, rowData, event, target);

            if (that.editing || that.drillDown) {
                if (!dataField || cell.classList.contains('no-edit')) {
                    return;
                }

                clearTimeout(that._dblclickObject.timeout);
                that._dblclickObject.numberOfClicks++;

                if (that._dblclickObject.numberOfClicks === 2) {
                    that._dblclickObject.numberOfClicks = 0;

                    that._beginEdit({
                        rowElement: row,
                        cell: cell,
                        dataField: dataField
                    });
                }
                else {
                    that._dblclickObject.timeout = setTimeout(function () {
                        that._dblclickObject.numberOfClicks = 0;
                    }, 300);
                }
            }

            if (that.rowDetailTemplate && event.target.closest('.detail-checkbox') && cell.classList.contains('smart-table-detail-row')) {
                that._toggleRowDetail(row, parsedId);
            }

            if (event.target.hasAttribute('action')) {
                const action = event.target.getAttribute('action');
                that._processAction(action, row, parsedId);
            }

            return;
        }


        const selectAllCheckbox = target.closest('.smart-table-select-all');

        if (selectAllCheckbox) {
            if (that.selectionMode === 'one') {
                return;
            }

            that._selectAllCheckboxClickHandler(selectAllCheckbox);
            return;
        }

        const clearFilterButton = target.closest('.smart-table-filter-row smart-button');

        if (clearFilterButton) {
            const filterValueInput = clearFilterButton.parentElement.firstElementChild,
                filterConditionInput = clearFilterButton.parentElement.children[1];

            filterValueInput.value = '';

            if (['EMPTY', 'NOT_EMPTY', 'NULL', 'NOT_NULL'].indexOf(filterConditionInput.$.input.dataValue) !== -1) {
                filterConditionInput.value = that.localize(filterValueInput.type === 'text' ? 'CONTAINS' : 'EQUAL');
                delete filterConditionInput.$.input.dataValue;
            }

            that._applyRowFilters();
            return;
        }
    }

    _processAction(action, row, rowId) {
        const that = this;

        switch (action) {
            case 'close':
                that.endRowEdit(rowId);
                break;

            default: {
                that.endRowEdit(rowId);
                break;
            }
        }
    }

    /**
    * Toggles row detail.
    */
    _toggleRowDetail(row, id) {
        const that = this,
            detailRow = row.nextElementSibling,
            detailCheckBox = row.querySelector('.detail-checkbox');

        if (detailCheckBox) {
            detailCheckBox.setAttribute('aria-checked', !detailRow || !detailRow.classList.contains('smart-table-row-detail'))
        }

        that._toggleRowDetailVirtualization(row, id);

    }

    _hideRowDetail(row, id) {
        const that = this,
            detailRow = row.nextElementSibling,
            detailCheckBox = row.querySelector('.detail-checkbox');

        if (detailCheckBox) {
            detailCheckBox.removeAttribute('aria-checked');
        }

        that._hideRowDetailVirtualization(row, id);
    }

    /**
     * Refresh the Tickets Container Scroll according to the Timeline
     * @returns 
     */
    _refreshTicketScroll() {
        const that = this,
            timeline = that._timeline,
            columnIndex = that.columns.toArray().findIndex(col => col.dataField === 'timeline');

        if (columnIndex < 0 || !timeline) {
            return
        }

        const rows = that._rowsInDOM,
            selectionCoefficient = Number(that.selection),
            detailCheckboxCoefficient = Number(!!that.rowDetailTemplate);

        for (let i = 0, max = rows.length; i < max; i += 1) {
            const td = rows[i].children[columnIndex + selectionCoefficient + detailCheckboxCoefficient];

            if (td && td.getAttribute('data-field') === 'timeline') {
                td.firstElementChild.lastElementChild.scrollLeft = timeline.scrollView.scrollLeft;
            }
        }
    }

    /**
    * resize handler.
    */
    _resizeHandler() {
        const that = this;

        clearTimeout(that._resizeTimeout);

        const updateStickyHeader = () => {
            if (Smart.Utilities.Core.Browser.Firefox && that.virtualization) {
                const scrollbarValue = that.$.virtualizationContainer.$.verticalScrollBar.value;
                that._onVerticalChange({ detail: { value: scrollbarValue } }, true);
            }
        }

        updateStickyHeader();
        that._resizeTimeout = setTimeout(function () {
            if (that.columnGroups &&
                that._columns.some(col => col.responsivePriority && col.responsivePriority > 1)) {
                that.refresh();
                return;
            }

            //Refresh the timeline and tickets
            that._recycleTimeline();
            that._refreshScrollButtons();
            that._recycleRowTickets();

            const currentWidth = that.offsetWidth,
                currentHeight = that.offsetHeight;

            if (that.virtualization && that._cachedHeight !== currentHeight) {
                const fractionOfMax = that._getFractionOfMax();

                that._refreshDataRows();
                that._setFractionOfMax(fractionOfMax);
                that._cachedWidth = currentWidth;
                that._cachedHeight = currentHeight;
                updateStickyHeader();
                return;
            }

            that._handleAutoSizeMode();
            that._handleFrozenColumnPositions();
            that._refreshHorizontalScrollbar();
            that._cachedWidth = currentWidth;
            that._cachedHeight = currentHeight;
            updateStickyHeader();
        }, 75);
    }

    /**
     * Recycles the Tickets
     * @returns 
     */
    _recycleRowTickets() {
        const that = this,
            columnIndex = that.columns.toArray().findIndex(col => col.dataField === 'timeline');

        if (columnIndex < 0) {
            return
        }

        const rows = that._rowsInDOM,
            column = that.columns[columnIndex],
            selectionCoefficient = Number(that.selection),
            detailCheckboxCoefficient = Number(!!that.rowDetailTemplate);

        for (let i = 0, max = rows.length; i < max; i += 1) {
            const rowObject = rows[i],
                td = rows[i].children[columnIndex + selectionCoefficient + detailCheckboxCoefficient];

            if (td && td.getAttribute('data-field') === 'timeline') {
                that._refreshTickets(that._formatCellValue(rowObject, column, td), td, rowObject.data ? rowObject.data.$.id : undefined);
            }
        }
    }

    /**
    * Updates visible rows (when virtualization is enabled).
    */
    _updateVisibleRows(startOfView, startOfData, forceUpdate) {
        const that = this,
            startCoefficient = startOfData - startOfView,
            numberOfVirtualRows = that._rowsVirtual.length,
            tableSettings = { hierarchical: that.hasAttribute('hierarchy'), selection: that.selection };
        let focusedCellInfo;

        forceUpdate = forceUpdate &&
            (tableSettings.hierarchical || !!that.rowDetailTemplate);

        if (tableSettings.hierarchical && !that.grouping) {
            forceUpdate = true;
        }

        if (that._sortColumns && that._sortColumns.length > 0) {
            forceUpdate = true;
        }

        if (that._focusedCell) {
            const focusedCellParentRow = that._focusedCell.parentElement;

            if (focusedCellParentRow) {
                focusedCellInfo = {
                    rowId: focusedCellParentRow.data.$.id,
                    dataField: that._focusedCell.getAttribute('data-field')
                };
            }
        }

        for (let i = 0; i < startOfView; i++) {
            const tr = that._rowsInDOM[i];
            let newIndex = startCoefficient + i;

            if (newIndex >= 0 && (forceUpdate || tr.data.$.index !== newIndex || tr.data.detail)) {
                that._updateVisibleRow(tr, newIndex, tableSettings);
            }

            tr.setAttribute('unused', '');
            tr.setAttribute('aria-hidden', true);
            tr.removeAttribute('edit');
            tr.classList.remove('smart-table-row-detail');
        }

        for (let i = startOfView; i < that._rowsInDOM.length; i++) {
            const tr = that._rowsInDOM[i],
                newIndex = startCoefficient + i;

            if (newIndex < numberOfVirtualRows) {
                tr.removeAttribute('unused');
                tr.removeAttribute('aria-hidden');

                if (forceUpdate || tr.data.$.index !== newIndex || tr.data.parent) {
                    that._updateVisibleRow(tr, newIndex, tableSettings);
                }
            }
            else {
                tr.setAttribute('aria-hidden', true);
            }
        }

        that._start = { view: startOfView, data: startOfData };
        that._focusEquivalentCell(focusedCellInfo);
    }

    /**
     * Updates a visible row with new data (when virtualization is enabled).
     */
    _updateVisibleRow(rowElement, newIndex, tableSettings) {
        const that = this,
            columns = that._columns.filter(col => col.visible);

        if (!columns.length) {
            return
        }

        const rowObject = that._rowsVirtual[newIndex],
            data = rowObject.data,
            id = data.$.id,
            selectionCoefficient = Number(tableSettings.selection),
            detailCheckboxCoefficient = Number(!!that.rowDetailTemplate),
            update = !data.groupDataField && rowElement.firstElementChild && !rowElement.firstElementChild.classList.contains('group-header') &&
                !rowObject.detail && !rowElement.classList.contains('smart-table-row-detail') && rowElement.childElementCount > 1,
            rowEditing = that._rowEditing;

        rowElement.className = '';

        if (update) {
            const attributes = rowElement.attributes;

            while (attributes.length > 0) {
                rowElement.removeAttribute(attributes[0].name);
            }
        }
        else {
            /* To do perf. test.
            while (rowElement.childNodes.length > 0) {
                rowElement.childNodes[0].remove();
            }*/
            rowElement.innerHTML = '';
        }

        if (data.groupDataField) {
            that._createGroupHeader(data, undefined, rowElement);
            return;
        }

        if (rowObject.detail) {
            that._createRowDetailOnDemand(rowElement, rowObject, data);
            return;
        }

        rowElement.setAttribute('aria-rowindex', data.$.index + 1);

        that._processAndRenderSelection(data, rowElement, tableSettings.hierarchical, update);

        that._processAndRenderDetails(data, rowElement, tableSettings.hierarchical, update);

        rowElement.data = data;
        rowElement.setAttribute('row-id', id);

        for (let j = 0; j < columns.length; j++) {
            const column = columns[j];
            let td = update ? rowElement.children[j + selectionCoefficient + detailCheckboxCoefficient] :
                document.createElement('td');
            let isNewCell = false;

            if (td && ((that.grouping && !data.groupDataField) || (that.dataSource && that.dataSource.groupBy && that.dataSource.groupBy.length > 0))) {
                td.removeAttribute('colspan');
            }

            if (!td) {
                if (update) {
                    td = document.createElement('td');
                    isNewCell = true;
                    rowElement.appendChild(td);
                }
                else {
                    continue;
                }
            }

            if (update) {
                td.className = '';
            }

            td.setAttribute('data-field', column.dataField);
            td.setAttribute('aria-colindex', j + 1 + selectionCoefficient + detailCheckboxCoefficient);

            let value = that._formatCellValue(rowObject, column, td, column.dataField === 'progress' ?
                { completed: data[column.dataField], total: Array.isArray(data.timeline) ? data.timeline.length : 0 } : undefined);

            let isTreeCell;

            if (isNewCell) {
                that._setCellContent(td, value, isTreeCell, id);
            }

            if (column.freeze) {
                td.classList.add('freeze-' + (column.freeze === 'far' ? 'far' : 'near'));
            }

            if (column.responsivePriority) {
                td.classList.add('priority-' + column.responsivePriority);
            }

            if (column.allowEdit === false) {
                td.classList.add('no-edit');
                td.setAttribute('aria-readonly', true);
            }
            else {
                td.removeAttribute('aria-readonly');
            }

            if (j === 0 && tableSettings.hierarchical) {
                const childrenFiltered = that._areChildrenFiltered(data);

                if (!data.leaf) {
                    td.classList.add('tree-cell');
                    isTreeCell = childrenFiltered;
                }

                if (data.level) {
                    td.classList.add('outline-level-' + Math.min(data.level, 10));

                    if (data.leaf || !childrenFiltered) {
                        td.classList.add('tree-leaf');
                    }
                }
            }

            if (column.templateElement) {
                if (isNewCell && that.grouping) {
                    const templateElement = column.templateElement + '';
                    const templateChild = td;

                    if (templateElement.indexOf('{{value}}') >= 0) {
                        templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                    }
                    else {
                        templateChild.innerHTML = templateElement;
                    }
                }
                else if (tableSettings.hierarchical && that.grouping && data.leaf) {
                    const templateElement = column.templateElement + '';
                    const templateChild = td;

                    const shouldUpdate = templateChild.firstElementChild && templateChild.firstElementChild.hasAttribute('ng-version') ? false : true;

                    if (shouldUpdate) {
                        if (templateChild.firstElementChild && !templateChild.firstElementChild.classList.contains('.smart-table-cell-template')) {
                            const html = templateChild.innerHTML;
                            const template = templateElement.substring(templateElement.lastIndexOf('</'));
                            const updateHTML = template !== html.substring(html.lastIndexOf('</'));

                            if (updateHTML) {
                                if (templateElement.indexOf('{{value}}') >= 0) {
                                    templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                                }
                                else {
                                    templateChild.innerHTML = templateElement;
                                }
                            }
                        }
                    }
                }

                if (isTreeCell && !td.querySelector('.hierarchy-arrow')) {
                    that._setCellContent(td, value, isTreeCell, id);

                    const templateChild = td.children[0].children[1];
                    const templateElement = column.templateElement + '';

                    if (templateElement.indexOf('{{value}}') >= 0) {
                        templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                    }
                    else {
                        templateChild.innerHTML = templateElement;
                    }
                }
                else if (tableSettings.hierarchical && !that.grouping && data.leaf) {
                    const templateChild = td;
                    const shouldUpdate = templateChild.firstElementChild && templateChild.firstElementChild.hasAttribute('ng-version') ? false : true;

                    if (shouldUpdate) {
                        that._setCellContent(td, value, false, id);

                        const templateElement = column.templateElement + '';

                        if (templateElement.indexOf('{{value}}') >= 0) {
                            templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                        }
                        else {
                            templateChild.innerHTML = templateElement;
                        }
                    }
                }

                if (!update) {
                    const templateElement = column.templateElement + '';

                    const templateChild = !isTreeCell ? td : td.children[0].children[1];

                    if (templateElement.indexOf('{{value}}') >= 0) {
                        templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                    }
                    else {
                        templateChild.innerHTML = templateElement;
                    }
                }

                let templateFirstChild = !isTreeCell ? td.firstElementChild : td.children[0].children[1].firstElementChild;

                if (!templateFirstChild && tableSettings.hierarchical) {
                    that._setCellContent(td, value, false, id);

                    const templateChild = td;
                    const templateElement = column.templateElement + '';

                    if (templateElement.indexOf('{{value}}') >= 0) {
                        templateChild.innerHTML = templateElement.replace(/{{value}}/ig, '<div class="smart-table-cell-template">' + data[column.dataField] + '</div>');
                    }
                    else {
                        templateChild.innerHTML = templateElement;
                    }

                    templateFirstChild = !isTreeCell ? td.firstElementChild : td.children[0].children[1].firstElementChild;
                }

                if (column.templateElementSettings) {
                    column.templateElementSettings(data[column.dataField], id, templateFirstChild);
                }
                else if (column.templateElement.indexOf('{{value}}') >= 0) {
                    const templateElement = td.querySelector('.smart-table-cell-template');

                    templateElement.textContent = data[column.dataField];
                }
            }
            else {
                if (column.formatFunction) {
                    that._setCellContent(td, value, isTreeCell, id);
                }
                else {
                    that._setCellContent(td, value, isTreeCell, id);
                }
            }

            if (that.onCellRender) {
                that.onCellRender(data, column.dataField, data[column.dataField], td);
            }

            that._applyConditionalFormattingToCell(td, column.dataField, data.$.index);

            if (!update) {
                rowElement.appendChild(td);
            }
        }

        if (that._disabledSelection.indexOf(id) !== -1) {
            rowElement.removeAttribute('aria-selected');
            rowElement.classList.add('disable-select');

            if (tableSettings.selection) {
                rowElement.firstElementChild.firstElementChild.setAttribute('aria-disabled', true);
            }
        }

        if (data.expanded) {
            if (that._expandedIds.indexOf(id) === -1) {
                that._expandedIds.push(id);
            }

            rowElement.setAttribute('aria-expanded', true);
            rowElement.classList.add('expanded');
        }
        else if (data.expanded === false) {
            rowElement.setAttribute('aria-expanded', false);
        }

        if (data.level && that._isCollapsed(data)) {
            rowElement.setAttribute('aria-hidden', true);
            rowElement.classList.add('collapsed', 'smart-hidden');
        }

        //TODO check unsued
        if (rowEditing && rowEditing.indexOf(id) > -1) {
            rowElement.setAttribute('edit', '');
        }
        else {
            rowElement.removeAttribute('edit');
        }
    }
})